# sage.doctest: optional - fricas
r"""
Interface to FriCAS

FriCAS is a free GPL-compatible (modified BSD license) general
purpose computer algebra system based on Axiom.  The FriCAS
website can be found at http://fricas.sourceforge.net/.

AUTHORS:

- Mike Hansen (2009-02): Split off the FriCAS interface from
  the Axiom interface.

- Martin Rubey, Bill Page (2016-08): Completely separate from Axiom,
  implement more complete translation from FriCAS to SageMath types.

- Martin Rubey, Chenxin Zhong (2026-04): implement proper translation from
  FriCAS to SageMath types.

EXAMPLES::

    sage: fricas('3 * 5')
    15
    sage: a = fricas(3) * fricas(5); a
    15

The type of a is :class:`FriCASElement`, i.e., an element of the
FriCAS interpreter::

    sage: type(a)
    <class 'sage.interfaces.fricas.FriCASElement'>
    sage: a.parent()
    FriCAS

The underlying FriCAS type of a is also available, via the type
method::

    sage: a.typeOf()
    PositiveInteger

FriCAS objects are normally displayed using "ASCII art"::

    sage: fricas(2/3)
      2
      -
      3
    sage: fricas('x^2 + 3/7')
       2   3
      x  + -
           7

Functions defined in FriCAS are available as methods of the :class:`fricas<FriCAS>` object::

    sage: F = fricas.factor('x^5 - y^5'); F
               4      3    2 2    3     4
    - (y - x)(y  + x y  + x y  + x y + x )
    sage: type(F)
    <class 'sage.interfaces.fricas.FriCASElement'>
    sage: F.typeOf()
    Factored(Polynomial(Integer))

We can also create a FriCAS polynomial and apply the function
``factor`` from FriCAS.  The notation ``f.factor()`` is consistent
with how the rest of SageMath works::

    sage: f = fricas('x^5 - y^5')
    sage: f^2
     10      5 5    10
    y   - 2 x y  + x
    sage: f.factor()
               4      3    2 2    3     4
    - (y - x)(y  + x y  + x y  + x y + x )

For many FriCAS types, translation to an appropriate SageMath type is
available::

    sage: f.factor().sage()
    (-1) * (-x + y) * (x^4 + x^3*y + x^2*y^2 + x*y^3 + y^4)

Control-C interruption works well with the FriCAS interface. For
example, try the following sum but with a much bigger range, and hit
control-C::

    sage:  f = fricas('(x^5 - y^5)^10000')                                      # not tested
    Interrupting FriCAS...
    ...
    KeyboardInterrupt: Ctrl-c pressed while running FriCAS

Let us demonstrate some features of FriCAS.  FriCAS can guess a
differential equation for the generating function for integer
partitions::

    sage: fricas("guessADE([partition n for n in 0..40], homogeneous==4)")
    [
      [
          n
        [x ]f(x):
               2    3 (iv)           2    2 ,              3  ,,,
              x f(x) f    (x) + (20 x f(x) f (x) + 5 x f(x) )f   (x)
    <BLANKLINE>
            +
                    2    2 ,,   2
              - 39 x f(x) f  (x)
    <BLANKLINE>
            +
                   2     ,   2            2 ,            3  ,,         2 ,   4
              (12 x f(x)f (x)  - 15 x f(x) f (x) + 4 f(x) )f  (x) + 6 x f (x)
    <BLANKLINE>
            +
                        ,   3          2 ,   2
              10 x f(x)f (x)  - 16 f(x) f (x)
    <BLANKLINE>
          =
            0
        ,
                         2      3      4
       f(x) = 1 + x + 2 x  + 3 x  + O(x )]
      ]

FriCAS can solve linear ordinary differential equations::

    sage: fricas.set("y", "operator y")
    sage: fricas.set("deq", "x^3*D(y x, x, 3) + x^2*D(y x, x, 2) - 2*x*D(y x, x) + 2*y x - 2*x^4")
    sage: fricas.set("sol", "solve(deq, y, x)"); fricas("sol")
                   5       3       2
                  x  - 10 x  + 20 x  + 4
    [particular = ----------------------,
                           15 x
                 3      2       3       3      2
              2 x  - 3 x  + 1  x  - 1  x  - 3 x  - 1
     basis = [---------------, ------, -------------]]
                     x            x          x

    sage: fricas("sol.particular").sage()
    1/15*(x^5 - 10*x^3 + 20*x^2 + 4)/x
    sage: fricas("sol.basis").sage()
    [(2*x^3 - 3*x^2 + 1)/x, (x^3 - 1)/x, (x^3 - 3*x^2 - 1)/x]
    sage: fricas.eval(")clear values y deq sol")
    ''

FriCAS can expand expressions into series::

    sage: x = var('x'); ex = sqrt(cos(x)); a = fricas(ex).series(x=0); a
        1  2    1  4    19   6     559   8     29161    10      11
    1 - - x  - -- x  - ---- x  - ------ x  - --------- x   + O(x  )
        4      96      5760      645120      116121600

    sage: a.coefficients()[38].sage()
    -29472026335337227150423659490832640468979/274214482066329363682430667508979749984665600000000

    sage: ex = sqrt(atan(x)); a = fricas(ex).series(x=0); a
     1      5        9
     -      -        -
     2   1  2    31  2      6
    x  - - x  + --- x  + O(x )
         6      360

    sage: a.coefficient(9/2).sage()
    31/360

    sage: x = fricas("x::TaylorSeries Fraction Integer")
    sage: y = fricas("y::TaylorSeries Fraction Integer")
    sage: 2*(1+2*x+sqrt(1-4*x)-2*x*y).recip()
                  2       3     2 2      3       4        4        5
      1 + (x y + x ) + 2 x  + (x y  + 2 x y + 6 x ) + (4 x y + 18 x )
    +
        3 3      4 2       5        6        5 2       6         7
      (x y  + 3 x y  + 13 x y + 57 x ) + (6 x y  + 40 x y + 186 x )
    +
        4 4      5 3       6 2        7         8
      (x y  + 4 x y  + 21 x y  + 130 x y + 622 x )
    +
          6 3       7 2        8          9
      (8 x y  + 66 x y  + 432 x y + 2120 x )
    +
        5 5      6 4       7 3        8 2         9          10
      (x y  + 5 x y  + 30 x y  + 220 x y  + 1466 x y + 7338 x  ) + O(11)

FriCAS does some limits right::

    sage: x = var('x'); ex = x^2*exp(-x)*Ei(x) - x; fricas(ex).limit(x=oo)
    1
"""

###########################################################################
#       Copyright (C) 2008 Mike Hansen <mhansen@gmail.com>
#                     2007 Bill Page
#                     2006 William Stein <wstein@gmail.com>
#
#  Distributed under the terms of the GNU General Public License (GPL)
#  The full text of the GPL is available at:
#
#                  https://www.gnu.org/licenses/
###########################################################################

import re
import os

import sage.interfaces.abc

from importlib.resources import files
from sage.env import DOT_SAGE, LOCAL_IDENTIFIER
from sage.interfaces.tab_completion import ExtraTabCompletion
from sage.interfaces.expect import Expect, ExpectElement, FunctionElement, ExpectFunction
from sage.misc.instancedoc import instancedoc
from sage.rings.integer_ring import ZZ
from sage.rings.rational_field import QQ
from sage.misc.lazy_import import lazy_import
lazy_import('sage.symbolic.expression', ['register_symbol'])
lazy_import('sage.symbolic.constants', ['I', 'e', 'pi'])

FRICAS_SINGLE_LINE_START = 3  # where output starts when it fits next to the line number
FRICAS_MULTI_LINE_START = 2   # and when it doesn't
FRICAS_LINE_LENGTH = 80       # length of a line, should match the line length in sage
# the following messages have, unfortunately, no markup.
FRICAS_WHAT_OPERATIONS_STRING = r"Operations whose names satisfy the above pattern\(s\):"
FRICAS_ERROR_IN_LIBRARY_CODE = ">> Error detected within library code:"
FRICAS_SYSTEM_ERROR = ">> System error:"

# only the last command should be necessary to make the interface
# work, the other are optimizations.  Beware that lisp distinguishes
# between ' and ".
FRICAS_INIT_CODE = (
    ")set functions compile on",
    ")set message any          on",
    ")set message autoload     off",
    ")set message bottomup     off",
    ")set message dropmap      off",
    ")set message expose       off",
    ")set message file         off",
    ")set message frame        off",
    ")set message highlighting off",
    ")set message instant      off",
    ")set message insteach     off",
    ")set message interponly   off",
    ")set message prompt       step",
    ")set message selection    off",
    ")set message set          off",
    ")set message startup      on",
    ")set message storage      off",
    ")set message testing      off",
    ")set message time         off",
    ")set message type         off",
    ")set message void         off",
    ")set output length " + str(FRICAS_LINE_LENGTH),
    ")lisp (setf |$ioHook|"
    "            (lambda (x &optional args)"
    "              (when (member x '(|startAlgebraOutput| |endOfAlgebraOutput|"
    "                                |startKeyedMsg|      |endOfKeyedMsg|))"
    "               (prin1 x)"
    "               (princ #\\Newline))))")
# code (one-line!) executed after having set up the prompt
FRICAS_HELPER_CODE = (
    'sageprint(x:SExpression):String == ' +
    '(atom? x => (' +
    'float? x => return float(x)::String;' +
    'integer? x => return integer(x)::String;' +
    'string? x => return concat(["_"", string(x)::String, "_""])$String;' +
    'symbol? x => return string(symbol(x));' +
    'list? x => return "()");' +
    'S: List String := [sageprint y for y in destruct x];' +
    'R: String := new(1 + reduce(_+, [1 + #(s)$String for s in S], 0),' +
    'space()$Character);' +
    'copyInto!(R, "(", 1);' +
    'i := 2;' +
    'for s in S repeat'
    '(copyInto!(R, s, i); i := i + 1 + #(s)$String);' +
    'copyInto!(R, ")", i-1);' +
    'return R)')

FRICAS_LINENUMBER_OFF_CODE = ")lisp (setf |$IOindex| NIL)"
FRICAS_FIRST_PROMPT = r"\(1\) -> "
FRICAS_LINENUMBER_OFF_PROMPT = r"\(NIL\) -> "


class FriCAS(ExtraTabCompletion, Expect):
    """
    Interface to a FriCAS interpreter.
    """
    def __init__(self, name='fricas', command=None,
                 script_subdirectory=None, logfile=None,
                 server=None, server_tmpdir=None):
        """
        Create an instance of the FriCAS interpreter.

        TESTS::

            sage: a = FriCAS()
            sage: isinstance(loads(dumps(a)), FriCAS)
            True

        Check that :issue:`25174` is fixed::

            sage: fricas(I)
            %i

            sage: integrate(sin(x)*exp(I*x), x, -pi, 0, algorithm='fricas')
            1/2*I*pi

            sage: fricas(I*sin(x)).sage()
            I*sin(x)

            sage: fricas(I*x).sage()
            I*x
        """
        if command is None:
            from sage.features.fricas import FriCAS
            command = [FriCAS(), "-nosman"]

        eval_using_file_cutoff = 4096 - 5  # magic number from Expect._eval_line (there might be a bug)
        assert max(len(c) for c in FRICAS_INIT_CODE) < eval_using_file_cutoff
        self.__eval_using_file_cutoff = eval_using_file_cutoff
        self._COMMANDS_CACHE = '%s/%s_commandlist_cache.sobj' % (DOT_SAGE, name)
        # we run the init code in _start to avoid spurious output
        Expect.__init__(self,
                        name=name,
                        prompt=FRICAS_FIRST_PROMPT,
                        command=command,
                        script_subdirectory=script_subdirectory,
                        server=server,
                        server_tmpdir=server_tmpdir,
                        restart_on_ctrlc=False,
                        verbose_start=False,
                        init_code=[],
                        logfile=logfile,
                        eval_using_file_cutoff=eval_using_file_cutoff)

    def _start(self):
        """
        Start the FriCAS interpreter and switch off the linenumbers.

        EXAMPLES::

            sage: a = FriCAS()
            sage: a.is_running()
            False
            sage: a._start()
            sage: a.is_running()
            True
            sage: a.quit()
        """
        # setting the prompt properly is necessary for restarting FriCAS
        self._prompt = FRICAS_FIRST_PROMPT
        Expect._start(self)
        for line in FRICAS_INIT_CODE:
            self.eval(line, reformat=False)
        # switching off the line numbers also modified the prompt
        self._prompt = FRICAS_LINENUMBER_OFF_PROMPT
        self.eval(FRICAS_LINENUMBER_OFF_CODE, reformat=False)
        self.eval(FRICAS_HELPER_CODE, reformat=False)
        # compile fricas.spad
        in_path = files('sage.interfaces').joinpath("fricas.spad")
        out_path = os.path.join(DOT_SAGE, 'fricas')
        try:
            os.makedirs(out_path)
        except OSError:
            if not os.path.isdir(out_path):
                raise
        self.eval(f")cd {out_path}")
        self.eval(f")compile {in_path}")
        # register translations between SymbolicRing and FriCAS Expression
        self._register_symbols()

    def _install_hints(self):
        """
        Hints for installing FriCAS on your computer.

        EXAMPLES::

            sage: print(fricas._install_hints())
            In order...
        """
        return r"""
In order to use the FriCAS interface you need to have FriCAS installed.
You can either run 'sage -i fricas' to install FriCAS as an optional
package within SageMath, or install FriCAS separately, see
http://fricas.sourceforge.net.
"""

    def _quit_string(self):
        """
        Return the string used to quit FriCAS.

        EXAMPLES::

            sage: fricas._quit_string()
            ')quit'
            sage: a = FriCAS()
            sage: a.is_running()
            False
            sage: a._start()
            sage: a.is_running()
            True
            sage: a.quit()
            sage: a.is_running()
            False

        TESTS:

        Ensure that a new process is started after ``quit()``::

            sage: p = fricas.pid()
            sage: fricas.quit()
            sage: fricas.pid() == p
            False
        """
        return ')quit'

    def _commands(self):
        """
        Return a list of commands available. This is done by parsing the
        result of the first section of the output of ')what things'.

        EXAMPLES::

            sage: cmds = fricas._commands()
            sage: len(cmds) > 100
            True
            sage: '<' in cmds
            True
            sage: 'factor' in cmds
            True
        """
        output = self.eval(")what operations", reformat=False)
        m = re.search(FRICAS_WHAT_OPERATIONS_STRING + r"\n(.*)\n\|startKeyedMsg\|",
                      output, flags=re.DOTALL)
        l = m.groups()[0].split()
        return l

    def _tab_completion(self, verbose=True, use_disk_cache=True):
        """
        Return a list of all the commands defined in Fricas and optionally
        (per default) store them to disk.

        EXAMPLES::

            sage: c = fricas._tab_completion(use_disk_cache=False, verbose=False)
            sage: len(c) > 100
            True
            sage: 'factor' in c
            True
            sage: '**' in c
            False
            sage: 'upperCase?' in c
            False
            sage: 'upperCase_q' in c
            True
            sage: 'upperCase_e' in c
            True
        """
        try:
            return self.__tab_completion
        except AttributeError:
            import sage.misc.persist
            if use_disk_cache:
                try:
                    self.__tab_completion = sage.misc.persist.load(self._COMMANDS_CACHE)
                    return self.__tab_completion
                except OSError:
                    pass
            if verbose:
                print("\nBuilding %s command completion list (this takes" % self)
                print("a few seconds only the first time you do it).")
                print("To force rebuild later, delete %s." % self._COMMANDS_CACHE)
            v = self._commands()

            # process the commands to strip out things which are not
            # valid Python identifiers
            valid = re.compile('[^a-zA-Z0-9_]+')
            names = [x for x in v if valid.search(x) is None]

            # replace trailing ? with _q and trailing ! with _e
            names += [x[:-1] + "_q" for x in v if x.endswith("?")]
            names += [x[:-1] + "_e" for x in v if x.endswith("!")]

            self.__tab_completion = names
            if len(v) > 200:
                # Fricas is actually installed.
                sage.misc.persist.save(v, self._COMMANDS_CACHE)
            return names

    def _read_in_file_command(self, filename):
        """
        Return the FriCAS command to read the file ``filename``.

        INPUT:

        - ``filename`` -- string ending in '.input'.

        OUTPUT: string with the command for reading filename without output

        TESTS:

        Evaluate a rather long line::

            sage: len(fricas([i for i in range(600)]))  # indirect doctest
            600
        """
        if not filename.endswith('.input'):
            raise ValueError("the filename must end with .input")

        return ')read %s )quiet' % filename

    def _remote_tmpfile(self):
        """
        Return a remote tmpfile ending with ".input" used to buffer long
        command lines sent to FriCAS.
        """
        try:
            return self.__remote_tmpfile
        except AttributeError:
            self.__remote_tmpfile = self._remote_tmpdir() + "/interface_%s:%s.input" % (LOCAL_IDENTIFIER, self.pid())
            return self.__remote_tmpfile

# what I expect from FriCAS:

# 1.) in set(self, var, value)
#
# no markers:
# there could be some "debugging" output, as in fricas("guessADE([1,1,1,1], debug==true)")
#
# startKeyedMsg: an error happened
#
# 2.) in get(self, var)
#
# |startAlgebraOutput\|...|endOfAlgebraOutput\|
#
# 3.) I also need a routine to send a system command and get its output.

    def _check_errors(self, line, output):
        """
        Check whether output contains an error and, if so, raise it.

        INPUT:

        - ``line`` -- string that was sent to FriCAS

        - ``output`` -- string returned by FriCAS

        OUTPUT: none

        TESTS::

            sage: fricas.set("x", "[i fo83r i in 0..17]")   # indirect doctest
            Traceback (most recent call last):
            ...
            RuntimeError: An error occurred when FriCAS evaluated '[i fo83r i in 0..17]':
              Line   1: x:=[i fo83r i in 0..17];
                       ...A..........B
            ...Error  A: Missing mate.
            ...Error  B: syntax error at top level
            ...Error  B: Possibly missing a ]
               3 error(s) parsing

            sage: fricas.set("x", "something stupid")       # indirect doctest
            Traceback (most recent call last):
            ...
            RuntimeError: An error occurred when FriCAS evaluated 'something stupid':
               There are no library operations named something
                  Use HyperDoc Browse or issue
                                           )what op something
                  to learn if there is any operation containing " something " in its
                  name.
            <BLANKLINE>
               Cannot find a definition or applicable library operation named
                  something with argument type(s)
                                            Variable(stupid)
            <BLANKLINE>
                  Perhaps you should use "@" to indicate the required return type, or
                  "$" to specify which version of the function you need.
        """
        # otherwise there might be a message
        m = re.search(r"\|startKeyedMsg\|\n(.*)\n\|endOfKeyedMsg\|",
                      output, flags=re.DOTALL)
        if m:
            replacements = [('|startKeyedMsg|\n', ''),
                            ('|endOfKeyedMsg|', '')]
            for old, new in replacements:
                output = output.replace(old, new)
            raise RuntimeError("An error occurred when FriCAS evaluated '%s':\n%s" % (line, output))

        # or even an error
        if FRICAS_ERROR_IN_LIBRARY_CODE in output or FRICAS_SYSTEM_ERROR in output:
            raise RuntimeError("An error occurred when FriCAS evaluated '%s':\n%s" % (line, output))

    @staticmethod
    def _register_symbols():
        """
        Register translations between elements of the ``SymbolicRing`` and FriCAS ``Expression`` domain.

        This method is called from :meth:`_start`, to work around a
        circular import problem involving ``pi``.
        """
        from sage.calculus.functional import diff
        from sage.functions.log import dilog, lambert_w
        from sage.functions.trig import sin, cos, tan, cot, sec, csc, asin
        from sage.functions.hyperbolic import tanh, sinh, cosh, coth, sech, csch
        from sage.functions.other import abs
        from sage.functions.gamma import gamma
        from sage.functions.special import elliptic_e, elliptic_f
        from sage.misc.functional import symbolic_sum, symbolic_prod
        from sage.rings.infinity import infinity
        register_symbol(pi, {'fricas': 'pi'}, 0)  # %pi::INFORM is %pi, but (pi) also exists
        register_symbol(lambda: infinity, {'fricas': 'infinity'}, 0)  # %infinity::INFORM is (infinity)
        register_symbol(lambda: infinity, {'fricas': 'plusInfinity'}, 0)  # %plusInfinity::INFORM is (plusInfinity)
        register_symbol(lambda: -infinity, {'fricas': 'minusInfinity'}, 0)  # %minusInfinity::INFORM is (minusInfinity)
        register_symbol(cos, {'fricas': 'cos'})
        register_symbol(sin, {'fricas': 'sin'})
        register_symbol(tan, {'fricas': 'tan'})
        register_symbol(cot, {'fricas': 'cot'})
        register_symbol(sec, {'fricas': 'sec'})
        register_symbol(csc, {'fricas': 'csc'})
        register_symbol(tanh, {'fricas': 'tanh'})
        register_symbol(sinh, {'fricas': 'sinh'})
        register_symbol(cosh, {'fricas': 'cosh'})
        register_symbol(coth, {'fricas': 'coth'})
        register_symbol(sech, {'fricas': 'sech'})
        register_symbol(csch, {'fricas': 'csch'})
        register_symbol(gamma, {'fricas': 'Gamma'}, 1)
        register_symbol(gamma, {'fricas': 'Gamma'}, 2)
        register_symbol(lambda x, y: elliptic_e(asin(x), y), {'fricas': 'ellipticE'}, 2)
        register_symbol(lambda x, y: elliptic_f(asin(x), y), {'fricas': 'ellipticF'}, 2)
        register_symbol(lambda x, y: x + y, {'fricas': '+'}, 2)
        register_symbol(lambda x, y: x - y, {'fricas': '-'}, 2)
        register_symbol(lambda x, y: x * y, {'fricas': '*'}, 2)
        register_symbol(lambda x, y: x / y, {'fricas': '/'}, 2)
        register_symbol(lambda x, y: x ** y, {'fricas': '^'}, 2)
        register_symbol(lambda f, x: diff(f, x), {'fricas': 'D'}, 2)
        register_symbol(lambda x, y: x + y * I, {'fricas': 'complex'}, 2)
        register_symbol(lambda x: dilog(1 - x), {'fricas': 'dilog'}, 1)
        register_symbol(lambda z: lambert_w(z), {'fricas': 'lambertW'}, 1)
        register_symbol(abs, {'fricas': 'abs'}, 1)
        # construct occurs in the InputForm of hypergeometricF
        register_symbol(lambda *x: x, {'fricas': 'construct'}, -1)
        # the following is a hack to deal with
        # integrate(sin((x^2+1)/x),x)::INFORM giving
        # (integral (sin (/ (+ (^ x 2) 1) x)) (:: x Symbol))
        register_symbol(lambda x, y: x, {'fricas': '::'}, 2)

        def _convert_eval(f, a, b):
            # it might be that FriCAS also returns a two-argument
            # eval, where the second argument is a list of equations,
            # in which case this function needs to be adapted
            return f.subs({a: b})

        register_symbol(_convert_eval, {'fricas': 'eval'}, 3)

        def _convert_sum(x, y):
            v, seg = y.operands()
            a, b = seg.operands()
            return symbolic_sum(x, v, a, b)

        def _convert_prod(x, y):
            v, seg = y.operands()
            a, b = seg.operands()
            return symbolic_prod(x, v, a, b)

        register_symbol(_convert_sum, {'fricas': 'sum'}, 2)
        register_symbol(_convert_prod, {'fricas': 'product'}, 2)

        def explicitly_not_implemented(*args):
            raise NotImplementedError("the translation of the FriCAS Expression '%s' to sage is not yet implemented" % args)

        register_symbol(lambda *args: explicitly_not_implemented("rootOfADE"), {'fricas': 'rootOfADE'}, 2) # to be removed once we fully on FriCAS 1.3.10+
        register_symbol(lambda *args: explicitly_not_implemented("FEseries"), {'fricas': 'FEseries'}, 2)
        register_symbol(lambda *args: explicitly_not_implemented("rootOfRec"), {'fricas': 'rootOfRec'}, 2)

    def set(self, var, value):
        """
        Set a variable to a value in FriCAS.

        INPUT:

        - ``var``, ``value`` -- strings; the first representing a valid
          FriCAS variable identifier, the second a FriCAS expression

        OUTPUT: none

        EXAMPLES::

            sage: fricas.set('xx', '2')
            sage: fricas.get('xx')
            '2'
        """
        cmd = '%s%s%s;' % (var, self._assign_symbol(), value)
        output = self.eval(cmd, reformat=False)
        self._check_errors(value, output)

    def get(self, var):
        r"""
        Get the string representation of the value (more precisely, the
        OutputForm) of a variable or expression in FriCAS.

        If FriCAS cannot evaluate `var` an error is raised.

        EXAMPLES::

            sage: fricas.set('xx', '2')
            sage: fricas.get('xx')
            '2'
            sage: a = fricas('(1 + sqrt(2))^5')
            sage: fricas.get(a.name())
            '    +-+\n29 \\|2  + 41'
            sage: fricas.get('(1 + sqrt(2))^5')
            '    +-+\n29 \\|2  + 41'
            sage: fricas.new('(1 + sqrt(2))^5')
                +-+
            29 \|2  + 41
        """
        output = self.eval(str(var), reformat=False)
        # if there is AlgebraOutput we ask no more
        m = re.search(r"\|startAlgebraOutput\|\n(.*)\n\|endOfAlgebraOutput\|",
                      output, flags=re.DOTALL)
        if m:
            lines = m.groups()[0].split("\n")
            if max(len(line) for line in lines) < FRICAS_LINE_LENGTH:
                return "\n".join(line[FRICAS_SINGLE_LINE_START:] for line in lines)
            return "\n".join(line[FRICAS_MULTI_LINE_START:] for line in lines)

        self._check_errors(var, output)

    def get_string(self, var):
        """
        Return the value of a FriCAS string as a string, without checking
        that it is a string.

        TESTS:

        We test that strings are returned properly::

            sage: r = fricas.get_string('concat([concat(string(i)," ") for i in 0..299])')
            sage: r == " ".join(str(i) for i in range(300)) + ' '
            True

            sage: fricas.get_string('concat([string(1) for i in 1..5])') == "1"*5
            True

            sage: fricas.get_string('concat([string(1) for i in 1..10000])') == "1"*10000
            True

        A problem with leading space::

            sage: s = "unparse((-1234567890123456789012345678901234567890123456789012345678901234567890*n::EXPR INT)::INFORM)"
            sage: fricas.get_string(s)
            '(-1234567890123456789012345678901234567890123456789012345678901234567890)*n'

        Check that :issue:`25628` is fixed::

            sage: var("a b"); f = 1/(1+a*cos(x))
            (a, b)
            sage: lF = integrate(f, x, algorithm='fricas')
            sage: (diff(lF[0], x) - f).simplify_trig()
            0
            sage: (diff(lF[1], x) - f).simplify_trig()
            0
            sage: f = 1/(b*x^2+a); lF = integrate(f, x, algorithm='fricas'); lF
            [1/2*log((2*a*b*x + (b*x^2 - a)*sqrt(-a*b))/(b*x^2 + a))/sqrt(-a*b),
             arctan(sqrt(a*b)*x/a)/sqrt(a*b)]
            sage: (diff(lF[0], x) - f).simplify_trig()
            0
            sage: (diff(lF[1], x) - f).simplify_trig()
            0
        """
        # strip removes leading and trailing whitespace, after that
        # we can assume that the first and the last character are
        # double quotes
        return self.get(str(var)).replace("\n", "").strip()[1:-1]

    def _assign_symbol(self):
        """
        Return the symbol used for setting a variable in FriCAS.

        EXAMPLES::

            sage: fricas.set("x", "1");     # indirect doctest
            sage: fricas.get("x")
            '1'
            sage: fricas.eval(")cl val x")
            ''
        """
        return ":="

    def _equality_symbol(self):
        """
        Return the equality testing logical symbol in FriCAS.

        EXAMPLES::

            sage: a = fricas(x==6); a       # indirect doctest
            x = 6

        A warning::

            sage: fricas.set("x", 2);
            sage: a = fricas(x==6); a
            2 = 6
            sage: fricas.eval(")cl val x")
            ''
        """
        return "="

    def _true_symbol(self):
        """
        Return the string used for ``True`` in FriCAS.

        EXAMPLES::

            sage: str(fricas("(1=1)@Boolean")) == fricas._true_symbol()
            True
        """
        return "true"

    def _false_symbol(self):
        """
        Return the string used for ``False`` in FriCAS.

        EXAMPLES::

            sage: str(fricas("(1~=1)@Boolean")) == fricas._false_symbol()
            True
        """
        return "false"

    def _inequality_symbol(self):
        """
        Return the string used for False in FriCAS.

        EXAMPLES::

            sage: fricas(x!=0)          # indirect doctest
            true
        """
        return '~='

    def _repr_(self):
        """
        EXAMPLES::

            sage: fricas                # indirect doctest
            FriCAS
        """
        return "FriCAS"

    def __reduce__(self):
        """
        EXAMPLES::

            sage: FriCAS().__reduce__()
            (<function reduce_load_fricas at 0x...>, ())
            sage: f, args = _
            sage: f(*args)
            FriCAS
        """
        return reduce_load_fricas, tuple([])

    def eval(self, code, strip=True, synchronize=False, locals=None, allow_use_file=True,
             split_lines='nofile', reformat=True, **kwds):
        """
        Evaluate ``code`` using FriCAS.

        Except ``reformat``, all arguments are passed to
        :meth:`sage.interfaces.expect.Expect.eval`.

        INPUT:

        - ``reformat`` -- boolean; remove the output markers when True

        This can also be used to pass system commands to FriCAS.

        EXAMPLES::

            sage: fricas.set("x", "1783"); fricas("x")
            1783
            sage: fricas.eval(")cl val x");
            ''
            sage: fricas("x")
            x
        """
        output = Expect.eval(self, code, strip=strip,
                             synchronize=synchronize, locals=locals,
                             allow_use_file=allow_use_file, split_lines=split_lines,
                             **kwds)
        # we remove carriage returns (\r) to make parsing easier
        # they are sent depending on how fricas was invoked:
        # on linux, "fricas -nox -noclef" sends "\r\n" and "fricas -nosman" sends "\n"
        output = output.replace('\r', '')
        if reformat:
            replacements = [('|startAlgebraOutput|\n', ''),
                            ('|endOfAlgebraOutput|', ''),
                            ('|startKeyedMsg|\n', ''),
                            ('|endOfKeyedMsg|', '')]
            for old, new in replacements:
                output = output.replace(old, new)

        return output

    def _function_class(self):
        """
        Return the FriCASExpectFunction class.

        EXAMPLES::

            sage: fricas._function_class()
            <class 'sage.interfaces.fricas.FriCASExpectFunction'>
            sage: type(fricas.gcd)
            <class 'sage.interfaces.fricas.FriCASExpectFunction'>
        """
        return FriCASExpectFunction

    def _object_class(self):
        """
        EXAMPLES::

            sage: fricas._object_class()
            <class 'sage.interfaces.fricas.FriCASElement'>
            sage: type(fricas(2))
            <class 'sage.interfaces.fricas.FriCASElement'>
        """
        return FriCASElement

    def _function_element_class(self):
        """
        Return the FriCAS function element class.

        EXAMPLES::

            sage: fricas._function_element_class()
            <class 'sage.interfaces.fricas.FriCASFunctionElement'>
            sage: type(fricas(2).gcd)
            <class 'sage.interfaces.fricas.FriCASFunctionElement'>
        """
        return FriCASFunctionElement

    def console(self):
        """
        Spawn a new FriCAS command-line session.

        EXAMPLES::

            sage: fricas.console()                                              # not tested
                             FriCAS (AXIOM fork) Computer Algebra System
                                    Version: FriCAS 1.0.5
                     Timestamp: Thursday February 19, 2009 at 06:57:33
            -----------------------------------------------------------------------------
               Issue )copyright to view copyright notices.
               Issue )summary for a summary of useful system commands.
               Issue )quit to leave AXIOM and return to shell.
            -----------------------------------------------------------------------------
        """
        fricas_console()


@instancedoc
class FriCASElement(ExpectElement, sage.interfaces.abc.FriCASElement):
    """
    Instances of this class represent objects in FriCAS.

    Using the method :meth:`sage` we can translate some of them to
    SageMath objects:

    .. automethod:: _sage_
    """
    def __len__(self):
        """
        Return the length of a list.

        EXAMPLES::

            sage: v = fricas('[x^i for i in 0..5]')
            sage: len(v)
            6

        TESTS:

        Streams are not handled yet::

            sage: oh = fricas('[i for i in 1..]')
            sage: len(oh)
            Traceback (most recent call last):
            ...
            TypeError: ...
        """
        P = self._check_valid()
        l = P('#(%s)' % self._name)
        return l.sage()

    def __iter__(self):
        """
        Return an iterator over ``self``.

        EXAMPLES::

            sage: L = fricas([4,5,6])
            sage: list(L)
            [4, 5, 6]

        TESTS:

        Streams are not handled yet::

            sage: oh = fricas('[i for i in 1..]')
            sage: next(iter(oh))       # known bug
        """
        for i in range(len(self)):
            yield self[i]

    def __getitem__(self, n):
        """
        We implement the sage conventions here, translating to 0-based iterables.

        We do not check validity, since many objects in FriCAS are
        iterable, in particular Streams

        TESTS::

            sage: fricas("[1,2,3]")[0]
            1

        Negative indices do work::

            sage: fricas("[1,2,3]")[-1]
            3

            sage: fricas("[1,2,3]")[-2]
            2

        Invalid indices raise exceptions::

            sage: fricas("[1,2,3]")[3]
            Traceback (most recent call last):
            ...
            TypeError: An error occurred when FriCAS evaluated 'elt(...,...)':
            <BLANKLINE>
            >> Error detected within library code:
            index out of range

        And streams are ok too::

            sage: oh = fricas('[i for i in 1..]')
            sage: oh[4]
            5
        """
        n = int(n)
        P = self._check_valid()
        if n < -1:
            try:
                l = len(self)
            except TypeError:
                raise
            else:
                n += l
                if not (0 <= n < l):
                    raise IndexError("index out of range")
        # use "elt" instead of "." here because then the error
        # message is clearer
        if n == -1:
            return P.new("elt(%s,last)" % (self._name))
        return P.new("elt(%s,%s)" % (self._name, n + 1))

    def __int__(self):
        """
        TESTS::

            sage: int(fricas(2))
            2
        """
        return int(self.sage())

    def bool(self):
        """
        Coerce the expression into a boolean.

        EXAMPLES::

            sage: fricas("1=1").bool()
            True
            sage: fricas("1~=1").bool()
            False
        """
        P = self._check_valid()
        return P.new(self._name + "::Boolean").sage()

    def __bool__(self):
        """
        Check whether the expression is different from zero.

        EXAMPLES::

            sage: fricas(0).is_zero()   # indirect doctest
            True
        """
        P = self._check_valid()
        return not P.new("zero?(%s)" % self._name).sage()

    def __float__(self):
        """
        TESTS::

            sage: float(fricas(2))
            2.0
        """
        return float(self.sage())

    def _integer_(self, Z=None):
        """
        EXAMPLES::

            sage: ZZ(fricas('1'))
            1
        """
        return ZZ(self.sage())

    def _rational_(self):
        """
        EXAMPLES::

            sage: QQ(fricas('-1/2'))
            -1/2
        """
        return QQ(self.sage())

    def gen(self, n):
        """
        Return an error, since the n-th generator in FriCAS is not well defined.
        """
        raise NotImplementedError

    def _latex_(self):
        r"""
        EXAMPLES::

            sage: latex(fricas("sin(x+y)/exp(z)*log(1+%e)"))
            \frac{{{\log \left( {{e+1}} \right)} \  {\sin \left( {{y+x}} \right)}}}{{{e} ^{z}}}

            sage: latex(fricas("matrix([[1,2],[3,4]])"))
            \left[ \begin{array}{cc} 1 & 2 \\ 3 & 4...\end{array}...\right]

            sage: latex(fricas("integrate(sin(x+1/x),x)"))
            \int ^{\displaystyle x} {{\sin \left( {{\frac{{{{ \%...} ^{2}}+1}}{ \%...}}} \right)} \  {d \%...}}
        """
        replacements = [(r'\sp ', '^'),
                        (r'\sp{', '^{'),
                        (r'\sb ', '_'),
                        (r'\sb{', '_{')]
        P = self._check_valid()
        s = P.get_string("latex(%s)" % self._name)
        for old, new in replacements:
            s = s.replace(old, new)
        return s

    def _sage_(self):
        r"""
        Convert ``self`` to a Sage object.

        EXAMPLES:

        Floats::

            sage: fricas(2.1234).sage()
            2.12340000000000
            sage: _.parent()
            Real Field with 53 bits of precision
            sage: a = RealField(100)(pi)
            sage: fricas(a).sage()
            3.1415926535897932384626433833
            sage: _.parent()
            Real Field with 100 bits of precision
            sage: fricas(a).sage() == a
            True
            sage: fricas(2.0).sage()
            2.00000000000000
            sage: _.parent()
            Real Field with 53 bits of precision

        Algebraic numbers::

            sage: a = fricas('(1 + sqrt(2))^5'); a
                +-+
            29 \|2  + 41
            sage: b = a.sage(); b
            82.0121933088198?
            sage: b.radical_expression()
            29*sqrt(2) + 41

        Integers modulo n::

            sage: fricas("((42^17)^1783)::IntegerMod(5^(5^5))").sage() == Integers(5^(5^5))((42^17)^1783)
            True

        Matrices over a prime field::

            sage: fricas("matrix [[1::PF 3, 2],[2, 0]]").sage().parent()
            Full MatrixSpace of 2 by 2 dense matrices over Finite Field of size 3

        We can also convert FriCAS's polynomials to Sage polynomials::

            sage: a = fricas("x^2 + 1"); a.typeOf()
            Polynomial(Integer)
            sage: a.sage()
            x^2 + 1
            sage: _.parent()
            Multivariate Polynomial Ring in x over Integer Ring
            sage: fricas('x^2 + y^2 + 1/2').sage()
            x^2 + y^2 + 1/2
            sage: _.parent()
            Multivariate Polynomial Ring in x, y over Rational Field

            sage: fricas("1$Polynomial Integer").sage()
            1

            sage: fricas("x^2/2").sage()
            1/2*x^2

            sage: x = polygen(QQ, 'x')
            sage: fricas(x+3).sage()
            x + 3

            sage: f = fricas("(y^2+3)::UP(y, INT)").sage(); f
            y^2 + 3
            sage: f.parent()
            Univariate Polynomial Ring in y over Integer Ring

            sage: fricas("(y^2+sqrt 3)::UP(y, AN)").sage()
            y^2 + 1.732050807568878?

        Rational functions::

            sage: fricas("x^2 + 1/z").sage()
            (x^2*z + 1)/z

            sage: fricas("(1/(1-x))::FRAC MPOLY([x,y], INT)").sage()
            (-1)/(x - 1)

        Expressions::

            sage: fricas(pi).sage()
            pi

            sage: fricas("sin(x+y)/exp(z)*log(1+%e)").sage()
            e^(-z)*log(e + 1)*sin(x + y)

            sage: fricas("factorial(n)").sage()
            factorial(n)

            sage: fricas("integrate(sin(x+y), x=0..1)").sage()
            -cos(y + 1) + cos(y)

            sage: fricas("integrate(x*sin(1/x), x=0..1)").sage()
            'failed'

            sage: fricas("integrate(sin((x^2+1)/x),x)").sage()
            integral(sin((x^2 + 1)/x), x)

        Matrices::

            sage: fricas("matrix [[x^n/2^m for n in 0..5] for m in 0..3]").sage()
            [      1       x     x^2     x^3     x^4     x^5]
            [    1/2   1/2*x 1/2*x^2 1/2*x^3 1/2*x^4 1/2*x^5]
            [    1/4   1/4*x 1/4*x^2 1/4*x^3 1/4*x^4 1/4*x^5]
            [    1/8   1/8*x 1/8*x^2 1/8*x^3 1/8*x^4 1/8*x^5]

            sage: fricas("matrix([[2,3],[4,x+5]])").diagonal().sage()
            (2, x + 5)

            sage: fricas("[1,1/x]::DirectProduct(2, Fraction UnivariatePolynomial(x, Integer))").sage()
            (1, 1/x)

        Lists::

            sage: fricas("[2^n/x^n for n in 0..5]").sage()
            [1, 2/x, 4/x^2, 8/x^3, 16/x^4, 32/x^5]

            sage: fricas("[matrix [[i for i in 1..n]] for n in 0..5]").sage()
            [[], [1], [1 2], [1 2 3], [1 2 3 4], [1 2 3 4 5]]

        Error handling::

            sage: s = fricas.guessPade("[fibonacci i for i in 0..10]"); s
                n        x
            [[[x ]- ----------]]
                     2
                    x  + x - 1
            sage: s.sage()
            Traceback (most recent call last):
            ...
            NotImplementedError: the translation of the FriCAS Expression 'FEseries' to sage is not yet implemented

            sage: s = fricas("series(sqrt(1+x), x=0)"); s
                  1     1  2    1  3    5   4    7   5    21   6    33   7    429   8
              1 + - x - - x  + -- x  - --- x  + --- x  - ---- x  + ---- x  - ----- x
                  2     8      16      128      256      1024      2048      32768
            +
               715   9    2431   10      11
              ----- x  - ------ x   + O(x  )
              65536      262144

            sage: s.sage()
            Traceback (most recent call last):
            ...
            NotImplementedError: UnivariatePuiseuxSeries cannot be translated from FriCAS to SageMath yet

        TESTS:

        Strings::

            sage: fricas('["abc", "xyz"]').sage()
            ['abc', 'xyz']

        Unions::

            sage: f = fricas('[5::Union(INT, "failed")]')
            sage: f.sage()
            [5]

            sage: f = fricas('["failed"::Union(INT, "failed")]')
            sage: f.sage()
            ['failed']

            sage: f = fricas('"bbb"::Union(a: INT, b: "bbb")')
            sage: f.sage()
            'bbb'

        Records from the differential equation solver::

            sage: x = var('x')
            sage: y = function('y')(x)
            sage: f = fricas(diff(y,x) + y - 1).solve(y.operator(), x)
            sage: f
                                        - x
            [particular = 1, basis = [%e   ]]

            sage: f.sage()["particular"]
            1
            sage: f.sage()["basis"]
            [e^(-x)]

        Expressions::

            sage: f = fricas('integrate(sin(x^2), x)'); f
                       +---+
                       | 2
            fresnelS(x |--- )
                      \|%pi
            -----------------
                   +---+
                   | 2
                   |---
                  \|%pi
            sage: f.sage()
            1/2*sqrt(2)*sqrt(pi)*fresnel_sin(sqrt(2)*x/sqrt(pi))

        Check that :issue:`22525` is fixed::

            sage: l = [sin, cos, sec, csc, cot, tan, asin, acos, atan, acot, acsc, asec, arcsin, arccos, arctan, arccot, arccsc, arcsec]
            sage: [f(x)._fricas_().sage().subs(x=0.9) for f in l]
            [0.783326909627483,
             0.621609968270664,
             1.60872581046605,
             1.27660621345890,
             0.793551147842317,
             1.26015821755034,
             1.11976951499863,
             0.451026811796262,
             0.732815101786507,
             0.837981225008390,
             1.57079632679490 - 0.467145308103262*I,
             0.467145308103262*I,
             1.11976951499863,
             0.451026811796262,
             0.732815101786507,
             0.837981225008390,
             1.57079632679490 - 0.467145308103262*I,
             0.467145308103262*I]
            sage: l = [tanh, sinh, cosh, coth, sech, csch, asinh, acosh, atanh, acoth, asech, acsch, arcsinh, arccosh, arctanh, arccoth, arcsech, arccsch]
            sage: [f(x)._fricas_().sage().subs(x=0.9) for f in l]
            [0.716297870199024,
             1.02651672570818,
             1.43308638544877,
             1.39606725303001,
             0.697794641100332,
             0.974168247780004,
             0.808866935652782,
             0.451026811796262*I,
             1.47221948958322,
             1.47221948958322 - 1.57079632679490*I,
             0.467145308103262,
             0.957800449200672,
             0.808866935652782,
             0.451026811796262*I,
             1.47221948958322,
             1.47221948958322 - 1.57079632679490*I,
             0.467145308103262,
             0.957800449200672]

        Check that :issue:`23782` is fixed::

            sage: s = '((3*n^10-25*n^9+50*n^8+62*n^7-229*n^6-25*n^5+320*n^4-12*n^3-144*n^2)/11520)::EXPR INT'
            sage: fricas(s).sage()
            1/3840*n^10 - 5/2304*n^9 + 5/1152*n^8 + 31/5760*n^7 - 229/11520*n^6 - 5/2304*n^5 + 1/36*n^4 - 1/960*n^3 - 1/80*n^2

        Some checks for digamma and polygamma (:issue:`31853`)::

            sage: fricas.digamma(1.0)
            - 0.5772156649_0153286061
            sage: psi(1.0)
            -0.577215664901533
            sage: fricas.polygamma(1, 1.0)
            1.644934066848226...
            sage: psi(1, 1).n()
            1.64493406684823

            sage: var("w")
            w
            sage: fricas.laplace(log(x), x, w).sage()
            -(euler_gamma + log(w))/w
            sage: fricas(laplace(log(x), x, w)).sage()
            -(euler_gamma + log(w))/w

        Check that :issue:`25224` is fixed::

            sage: integrate(log(x)/(1-x),x,algorithm='fricas')
            dilog(-x + 1)
            sage: fricas(dilog(-x + 1))
            dilog(x)
            sage: dilog._fricas_()(1.0)
            1.6449340668_4822643647_24152
            sage: dilog(1.0)
            1.64493406684823

        Check that :issue:`25987` is fixed::

            sage: integrate(lambert_w(x), x, algorithm='fricas')
            (x*lambert_w(x)^2 - x*lambert_w(x) + x)/lambert_w(x)

        Check that :issue:`25838` is fixed::

            sage: F = function('f'); f = SR.var('f')
            sage: FF = fricas(F(f)); FF
            f(f)
            sage: FF.D(f).sage()
            diff(f(f), f)
            sage: bool(FF.D(f).integrate(f).sage() == F(f))
            True

        Check that :issue:`25602` is fixed::

            sage: r = fricas.integrate(72000/(1+x^5), x).sage()
            sage: abs(n(r.subs(x=5) - r.subs(x=3)) - 193.020947266210) <= 0.1
            True

            sage: var("a"); r = fricas.integrate(72000*a^8/(a^5+x^5), x).sage()
            a
            sage: abs(n(r.subs(a=1, x=5) - r.subs(a=1, x=3)) - 193.020947266268) <= 0.1
            True

        Check conversions of sums and products::

            sage: var("k, m, n")
            (k, m, n)
            sage: fricas("sum(1/factorial(k), k=1..n)").sage()
            sum(1/factorial(_...), _..., 1, n)
            sage: fricas("eval(sum(x/k, k=1..n), x=k)").sage()
            k*harmonic_number(n)
            sage: fricas("product(1/factorial(k), k=1..n)").sage()
            1/product(factorial(_...), _..., 1, n)

            sage: f = fricas.guess([sum(1/k, k,1,n) for n in range(10)])[0]; f
             n - 1
              --+       1
              >      -------
              --+    s   + 1
            s   = 0   10
             10

            sage: f.sage()
            harmonic_number(n)

            sage: f = fricas.guess([0, 1, 3, 9, 33])[0]; f
                    s  - 1
            n - 1    5
             --+    ++-++
             >       | |    p  + 2
             --+     | |     4
            s  = 0  p  = 0
             5       4

            sage: f.sage()
            sum(factorial(_... + 1), _..., 0, n - 1)

        Check that :issue:`26746` is fixed::

            sage: _ = var('x, y, z')
            sage: f = sin(x^2) + y^z
            sage: f.integrate(x, algorithm='fricas')
            1/2*sqrt(2)*sqrt(pi)*(sqrt(2)*x*y^z/sqrt(pi) + fresnel_sin(sqrt(2)*x/sqrt(pi)))

            sage: fricas(fresnel_sin(1))
            fresnelS(1)
            sage: fricas("fresnelS(1.0)")
            0.4382591473_9035476607_676

            sage: fricas(fresnel_cos(1))
            fresnelC(1)
            sage: fricas("fresnelC(1.0)")
            0.7798934003_7682282947_42

        Check that :issue:`17908` is fixed::

            sage: fricas(abs(x)).sage().subs(x=-1783)
            1783

        Check that :issue:`27310` is fixed::

            sage: fricas.set("F", "operator 'f")
            sage: fricas("eval(D(F(x,y), [x, y], [2, 1]), x=x+y)").sage()
            D[0, 0, 1](f)(x + y, y)

        Conversion of hypergeometric functions (:issue:`31298`)::

            sage: a,b,c = var("a b c")
            sage: A = hypergeometric([a, b], [c], x)
            sage: fricas(A).sage() - A
            0
            sage: fricas(A).D(x).sage() - diff(A, x)
            0

        Check that :issue:`31858` is fixed::

            sage: fricas.Gamma(3/2).sage()
            1/2*sqrt(pi)
            sage: fricas.Gamma(3/4).sage()
            gamma(3/4)
            sage: fricas.Gamma(3, 2).sage()
            gamma(3, 2)


        Check that :issue:`32133` is fixed::

            sage: var("y")
            y
            sage: f = fricas.zerosOf(y^4 + y + 1, y); f
                        +-----------------------------+
                        |       2                    2
                       \|- 3 %y1  - 2 %y0 %y1 - 3 %y0   - %y1 - %y0
            [%y0, %y1, --------------------------------------------,
                                             2
                +-----------------------------+
                |       2                    2
             - \|- 3 %y1  - 2 %y0 %y1 - 3 %y0   - %y1 - %y0
             ----------------------------------------------]
                                    2

            sage: f[1].sage()
            -1/2*sqrt(1/3)*sqrt((3*(1/18*I*sqrt(229)*sqrt(3) + 1/2)^(2/3) + 4)/(1/18*I*sqrt(229)*sqrt(3) + 1/2)^(1/3)) + 1/2*sqrt(-(1/18*I*sqrt(229)*sqrt(3) + 1/2)^(1/3) + 6*sqrt(1/3)/sqrt((3*(1/18*I*sqrt(229)*sqrt(3) + 1/2)^(2/3) + 4)/(1/18*I*sqrt(229)*sqrt(3) + 1/2)^(1/3)) - 4/3/(1/18*I*sqrt(229)*sqrt(3) + 1/2)^(1/3))
        """
        from sage.interfaces.fricas_translator import SEXParser, SEXPorter, SEXEvaluator, LazyParent
        P = self._check_valid()
        dom_str = P.get_string(f"sageprint(dom({self._name}))")
        dom = SEXParser(dom_str).parse()
        fun = SEXPorter(dom).export_call()
        obj_str = P.get_string(f"sageprint({fun}({self._name}))")
        obj = SEXParser(obj_str).parse()
        return SEXEvaluator(obj, LazyParent(dom)).eval()


@instancedoc
class FriCASFunctionElement(FunctionElement):
    def __init__(self, object, name):
        """
        Make FriCAS operation names valid python function identifiers.

        TESTS::

            sage: a = fricas('"Hello"')
            sage: a.upperCase_q
            upperCase?
            sage: a.upperCase_e
            upperCase!
            sage: a.upperCase_e()
            "HELLO"
        """
        if name.endswith("_q"):
            name = name[:-2] + "?"
        elif name.endswith("_e"):
            name = name[:-2] + "!"
        FunctionElement.__init__(self, object, name)


@instancedoc
class FriCASExpectFunction(ExpectFunction):
    def __init__(self, parent, name):
        """
        Translate the pythonized function identifier back to a FriCAS
        operation name.

        TESTS::

            sage: fricas.upperCase_q
            upperCase?
            sage: fricas.upperCase_e
            upperCase!
        """
        if name.endswith("_q"):
            name = name[:-2] + "?"
        elif name.endswith("_e"):
            name = name[:-2] + "!"
        ExpectFunction.__init__(self, parent, name)


fricas = FriCAS()


def reduce_load_fricas():
    """
    Return the FriCAS interface object defined in :mod:`sage.interfaces.fricas`.

    EXAMPLES::

        sage: from sage.interfaces.fricas import reduce_load_fricas
        sage: reduce_load_fricas()
        FriCAS
    """
    return fricas


def fricas_console():
    """
    Spawn a new FriCAS command-line session.

    EXAMPLES::

        sage: fricas_console()                                                  # not tested
                         FriCAS (AXIOM fork) Computer Algebra System
                                    Version: FriCAS 1.0.5
                     Timestamp: Thursday February 19, 2009 at 06:57:33
        -----------------------------------------------------------------------------
           Issue )copyright to view copyright notices.
           Issue )summary for a summary of useful system commands.
           Issue )quit to leave AXIOM and return to shell.
        -----------------------------------------------------------------------------
    """
    from sage.repl.rich_output.display_manager import get_display_manager
    if not get_display_manager().is_in_terminal():
        raise RuntimeError('Can use the console only in the terminal. Try %%fricas magics instead.')
    os.system('fricas -nox')


def __doctest_cleanup():
    """
    EXAMPLES::

        sage: from sage.interfaces.fricas import __doctest_cleanup
        sage: a = FriCAS()
        sage: two = a(2)
        sage: a.is_running()
        True
        sage: __doctest_cleanup()
        sage: a.is_running()
        False
    """
    import sage.interfaces.quit
    sage.interfaces.quit.expect_quitall()



# distutils: libraries = flint
# distutils: depends = flint/fmpz_mod_poly_factor.h

################################################################################
# This file is auto-generated by the script
#   SAGE_ROOT/tools/flint-autogen/flint_autogen.py.
# From the commit 3e2c3a3e091106a25ca9c6fba28e02f2cbcd654a
# Do not modify by hand! Fix and rerun the script instead.
################################################################################

from libc.stdio cimport FILE
from sage.libs.gmp.types cimport *
from sage.libs.mpfr.types cimport *
from sage.libs.flint.types cimport *

cdef extern from "flint_wrap.h":
    void fmpz_mod_poly_factor_init(fmpz_mod_poly_factor_t fac, const fmpz_mod_ctx_t ctx) noexcept
    void fmpz_mod_poly_factor_clear(fmpz_mod_poly_factor_t fac, const fmpz_mod_ctx_t ctx) noexcept
    void fmpz_mod_poly_factor_realloc(fmpz_mod_poly_factor_t fac, slong alloc, const fmpz_mod_ctx_t ctx) noexcept
    void fmpz_mod_poly_factor_fit_length(fmpz_mod_poly_factor_t fac, slong len, const fmpz_mod_ctx_t ctx) noexcept
    void fmpz_mod_poly_factor_set(fmpz_mod_poly_factor_t res, const fmpz_mod_poly_factor_t fac, const fmpz_mod_ctx_t ctx) noexcept
    void fmpz_mod_poly_factor_print(const fmpz_mod_poly_factor_t fac, const fmpz_mod_ctx_t ctx) noexcept
    void fmpz_mod_poly_factor_insert(fmpz_mod_poly_factor_t fac, const fmpz_mod_poly_t poly, slong exp, const fmpz_mod_ctx_t ctx) noexcept
    void fmpz_mod_poly_factor_concat(fmpz_mod_poly_factor_t res, const fmpz_mod_poly_factor_t fac, const fmpz_mod_ctx_t ctx) noexcept
    void fmpz_mod_poly_factor_pow(fmpz_mod_poly_factor_t fac, slong exp, const fmpz_mod_ctx_t ctx) noexcept
    bint fmpz_mod_poly_is_irreducible(const fmpz_mod_poly_t f, const fmpz_mod_ctx_t ctx) noexcept
    bint fmpz_mod_poly_is_irreducible_ddf(const fmpz_mod_poly_t f, const fmpz_mod_ctx_t ctx) noexcept
    bint fmpz_mod_poly_is_irreducible_rabin(const fmpz_mod_poly_t f, const fmpz_mod_ctx_t ctx) noexcept
    bint fmpz_mod_poly_is_irreducible_rabin_f(fmpz_t r, const fmpz_mod_poly_t f, const fmpz_mod_ctx_t ctx) noexcept
    bint _fmpz_mod_poly_is_squarefree(const fmpz * f, slong len, const fmpz_mod_ctx_t ctx) noexcept
    bint _fmpz_mod_poly_is_squarefree_f(fmpz_t fac, const fmpz * f, slong len, const fmpz_mod_ctx_t ctx) noexcept
    bint fmpz_mod_poly_is_squarefree(const fmpz_mod_poly_t f, const fmpz_mod_ctx_t ctx) noexcept
    bint fmpz_mod_poly_is_squarefree_f(fmpz_t fac, const fmpz_mod_poly_t f, const fmpz_mod_ctx_t ctx) noexcept
    bint fmpz_mod_poly_factor_equal_deg_prob(fmpz_mod_poly_t factor, flint_rand_t state, const fmpz_mod_poly_t pol, slong d, const fmpz_mod_ctx_t ctx) noexcept
    void fmpz_mod_poly_factor_equal_deg(fmpz_mod_poly_factor_t factors, const fmpz_mod_poly_t pol, slong d, const fmpz_mod_ctx_t ctx) noexcept
    void fmpz_mod_poly_factor_distinct_deg(fmpz_mod_poly_factor_t res, const fmpz_mod_poly_t poly, slong * const * degs, const fmpz_mod_ctx_t ctx) noexcept
    void fmpz_mod_poly_factor_distinct_deg_threaded(fmpz_mod_poly_factor_t res, const fmpz_mod_poly_t poly, slong * const * degs, const fmpz_mod_ctx_t ctx) noexcept
    void fmpz_mod_poly_factor_squarefree(fmpz_mod_poly_factor_t res, const fmpz_mod_poly_t f, const fmpz_mod_ctx_t ctx) noexcept
    void fmpz_mod_poly_factor(fmpz_mod_poly_factor_t res, const fmpz_mod_poly_t f, const fmpz_mod_ctx_t ctx) noexcept
    void fmpz_mod_poly_factor_cantor_zassenhaus(fmpz_mod_poly_factor_t res, const fmpz_mod_poly_t f, const fmpz_mod_ctx_t ctx) noexcept
    void fmpz_mod_poly_factor_kaltofen_shoup(fmpz_mod_poly_factor_t res, const fmpz_mod_poly_t poly, const fmpz_mod_ctx_t ctx) noexcept
    void fmpz_mod_poly_factor_berlekamp(fmpz_mod_poly_factor_t factors, const fmpz_mod_poly_t f, const fmpz_mod_ctx_t ctx) noexcept
    void _fmpz_mod_poly_interval_poly_worker(void * arg_ptr) noexcept
    void fmpz_mod_poly_roots(fmpz_mod_poly_factor_t r, const fmpz_mod_poly_t f, int with_multiplicity, const fmpz_mod_ctx_t ctx) noexcept
    int fmpz_mod_poly_roots_factored(fmpz_mod_poly_factor_t r, const fmpz_mod_poly_t f, int with_multiplicity, const fmpz_factor_t n, const fmpz_mod_ctx_t ctx) noexcept



"""
Root system data for (untwisted) type A affine
"""
#*****************************************************************************
#       Copyright (C) 2008-2009 Nicolas M. Thiery <nthiery at users.sf.net>,
#
#  Distributed under the terms of the GNU General Public License (GPL)
#                  http://www.gnu.org/licenses/
#*****************************************************************************

from .cartan_type import CartanType_standard_untwisted_affine, CartanType_simply_laced


class CartanType(CartanType_standard_untwisted_affine):
    def __init__(self, n):
        """
        EXAMPLES::

            sage: ct = CartanType(['A',4,1])
            sage: ct
            ['A', 4, 1]
            sage: ct._repr_(compact = True)
            'A4~'

            sage: ct.is_irreducible()
            True
            sage: ct.is_finite()
            False
            sage: ct.is_affine()
            True
            sage: ct.is_untwisted_affine()
            True
            sage: ct.is_crystallographic()
            True
            sage: ct.is_simply_laced()
            True
            sage: ct.classical()
            ['A', 4]
            sage: ct.dual()
            ['A', 4, 1]

            sage: ct = CartanType(['A', 1, 1])
            sage: ct.is_simply_laced()
            False
            sage: ct.dual()
            ['A', 1, 1]

        TESTS::

            sage: TestSuite(ct).run()
        """
        assert n >= 1
        CartanType_standard_untwisted_affine.__init__(self, "A", n)
        if n >= 2:
            self._add_abstract_superclass(CartanType_simply_laced)

    def _latex_(self):
        """
        Return a latex representation of ``self``.

        EXAMPLES::

            sage: ct = CartanType(['A',4,1])
            sage: latex(ct)
            A_{4}^{(1)}
        """
        return "A_{%s}^{(1)}" % self.n

    def dynkin_diagram(self):
        """
        Return the extended Dynkin diagram for affine type A.

        EXAMPLES::

            sage: a = CartanType(['A',3,1]).dynkin_diagram(); a                         # needs sage.graphs
             0
             O-------+
             |       |
             |       |
             O---O---O
             1   2   3
             A3~
            sage: a.edges(sort=True)                                                    # needs sage.graphs
            [(0, 1, 1),
             (0, 3, 1),
             (1, 0, 1),
             (1, 2, 1),
             (2, 1, 1),
             (2, 3, 1),
             (3, 0, 1),
             (3, 2, 1)]

            sage: a = DynkinDiagram(['A',1,1]); a                                       # needs sage.graphs
            O<=>O
            0   1
            A1~
            sage: a.edges(sort=True)                                                    # needs sage.graphs
            [(0, 1, 2), (1, 0, 2)]
        """
        from .dynkin_diagram import DynkinDiagram_class
        n = self.n
        g = DynkinDiagram_class(self)

        if n == 1:
            g.add_edge(0, 1, 2)
            g.add_edge(1, 0, 2)
        else:
            for i in range(1, n):
                g.add_edge(i, i+1)
            g.add_edge(0, 1)
            g.add_edge(0, n)
        return g

    def _latex_dynkin_diagram(self, label=None, node=None, node_dist=2):
        r"""
        Return a latex representation of the Dynkin diagram.

        EXAMPLES::

            sage: print(CartanType(['A',4,1])._latex_dynkin_diagram())
            \draw (0 cm,0) -- (6 cm,0);
            \draw (0 cm,0) -- (3.0 cm, 1.2 cm);
            \draw (3.0 cm, 1.2 cm) -- (6 cm, 0);
            \draw[fill=white] (0 cm, 0 cm) circle (.25cm) node[below=4pt]{$1$};
            \draw[fill=white] (2 cm, 0 cm) circle (.25cm) node[below=4pt]{$2$};
            \draw[fill=white] (4 cm, 0 cm) circle (.25cm) node[below=4pt]{$3$};
            \draw[fill=white] (6 cm, 0 cm) circle (.25cm) node[below=4pt]{$4$};
            \draw[fill=white] (3.0 cm, 1.2 cm) circle (.25cm) node[anchor=south east]{$0$};
            <BLANKLINE>
        """
        if label is None:
            label = lambda i: i
        if node is None:
            node = self._latex_draw_node
        if self.n == 1:
            ret = "\\draw (0, 0.1 cm) -- +(%s cm,0);\n" % node_dist
            ret += "\\draw (0, -0.1 cm) -- +(%s cm,0);\n" % node_dist
            ret += self._latex_draw_arrow_tip(0.33*node_dist-0.2, 0, 180)
            ret += self._latex_draw_arrow_tip(0.66*node_dist+0.2, 0, 0)
            ret += node(0, 0, label(0))
            ret += node(node_dist, 0, label(1))
            return ret
        rt_most = (self.n-1)*node_dist
        mid = 0.5 * rt_most
        ret = "\\draw (0 cm,0) -- (%s cm,0);\n" % rt_most
        ret += "\\draw (0 cm,0) -- (%s cm, 1.2 cm);\n" % mid
        ret += "\\draw (%s cm, 1.2 cm) -- (%s cm, 0);\n" % (mid, rt_most)
        for i in range(self.n):
            ret += node(i*node_dist, 0, label(i+1))
        ret += node(mid, 1.2, label(0), 'anchor=south east')
        return ret

    def ascii_art(self, label=None, node=None):
        """
        Return an ascii art representation of the extended Dynkin diagram.

        EXAMPLES::

            sage: print(CartanType(['A',3,1]).ascii_art())
            0
            O-------+
            |       |
            |       |
            O---O---O
            1   2   3

            sage: print(CartanType(['A',5,1]).ascii_art(label = lambda x: x+2))
            2
            O---------------+
            |               |
            |               |
            O---O---O---O---O
            3   4   5   6   7

            sage: print(CartanType(['A',1,1]).ascii_art())
            O<=>O
            0   1

            sage: print(CartanType(['A',1,1]).ascii_art(label = lambda x: x+2))
            O<=>O
            2   3
        """
        if label is None:
            label = lambda i: i
        if node is None:
            node = self._ascii_art_node
        n = self.n
        if n == 1:
            l0 = label(0)
            l1 = label(1)
            return "{}<=>{}\n{!s:4}{}".format(node(l0), node(l1), l0, l1)
        ret = "{}\n{}".format(label(0), node(label(0)))
        ret += "----"*(n-2) + "---+\n|" + "    "*(n-2) + "   |\n|" + "    "*(n-2) + "   |\n"
        ret += "---".join(node(label(i)) for i in range(1,n+1)) + "\n"
        ret += "".join("{!s:4}".format(label(i)) for i in range(1,n+1))
        return ret

    def dual(self):
        """
        Type `A_1^1` is self dual despite not being simply laced.

        EXAMPLES::

            sage: CartanType(['A',1,1]).dual()
            ['A', 1, 1]
        """
        return self

    def _default_folded_cartan_type(self):
        r"""
        Return the default folded Cartan type.

        In general, this just returns ``self`` in ``self`` with `\sigma` as
        the identity map.

        EXAMPLES::

            sage: CartanType(['A',1,1])._default_folded_cartan_type()
            ['A', 1, 1] as a folding of ['A', 3, 1]
            sage: CartanType(['A',3,1])._default_folded_cartan_type()
            ['A', 3, 1] as a folding of ['A', 3, 1]
        """
        from sage.combinat.root_system.type_folded import CartanTypeFolded
        if self.n == 1:
            return CartanTypeFolded(self, ['A', 3, 1], [[0,2], [1,3]])
        return CartanTypeFolded(self, self, [[i] for i in self.index_set()])



from sage.modular.pollack_stevens.space import PollackStevensModularSymbols
from sage.modular.pollack_stevens.distributions import Symk
from sage.modular.pollack_stevens.distributions import OverconvergentDistributions



import builtins
from typing import Any


class LaurentSeriesRingElement:
    def __init__(self, parent: Any, f: Any, n: int = 0) -> None: ...
    def __reduce__(self) -> tuple: ...
    def change_ring(self, R: Any) -> LaurentSeriesRingElement: ...
    def is_unit(self) -> bool: ...
    def is_zero(self) -> bool: ...
    def is_monomial(self) -> bool: ...
    def __bool__(self) -> bool: ...
    def _im_gens_(
        self, codomain: Any, im_gens: builtins.list[Any], base_map: Any | None = None
    ) -> LaurentSeriesRingElement: ...
    def _normalize(self) -> None: ...
    def _repr_(self) -> str: ...
    def verschiebung(self, n: int) -> LaurentSeriesRingElement: ...

    V = verschiebung

    def _latex_(self) -> str: ...
    def __hash__(self) -> int: ...
    def __getitem__(self, i: int | slice) -> Any: ...
    def __iter__(self) -> Any: ...
    def list(self) -> builtins.list[Any]: ...
    def coefficients(self) -> builtins.list[Any]: ...
    def residue(self) -> Any: ...
    def exponents(self) -> builtins.list[int]: ...
    def laurent_polynomial(self) -> Any: ...
    def lift_to_precision(
        self, absprec: int | None = None
    ) -> LaurentSeriesRingElement: ...
    def __setitem__(self, n: int, value: Any) -> None: ...
    def _unsafe_mutate(self, i: int, value: Any) -> None: ...
    def _add_(self, right_m: Any) -> LaurentSeriesRingElement: ...
    def _sub_(self, right_m: Any) -> LaurentSeriesRingElement: ...
    def add_bigoh(self, prec: int) -> LaurentSeriesRingElement: ...
    def O(self, prec: int) -> LaurentSeriesRingElement: ...
    def degree(self) -> int: ...
    def __neg__(self) -> LaurentSeriesRingElement: ...
    def _mul_(self, right_r: Any) -> LaurentSeriesRingElement: ...
    def _rmul_(self, c: Any) -> LaurentSeriesRingElement: ...
    def _lmul_(self, c: Any) -> LaurentSeriesRingElement: ...
    def __pow__(self, r: Any, dummy: Any) -> LaurentSeriesRingElement: ...
    def shift(self, k: int) -> LaurentSeriesRingElement: ...
    def __lshift__(self, k: int) -> LaurentSeriesRingElement: ...
    def __rshift__(self, k: int) -> LaurentSeriesRingElement: ...
    def truncate(self, n: int) -> LaurentSeriesRingElement: ...
    def truncate_laurentseries(self, n: int) -> LaurentSeriesRingElement: ...
    def truncate_neg(self, n: int) -> LaurentSeriesRingElement: ...
    def _div_(self, right_r: Any) -> LaurentSeriesRingElement: ...
    def common_prec(self, other: LaurentSeriesRingElement) -> int: ...
    def common_valuation(self, other: LaurentSeriesRingElement) -> int: ...
    def _richcmp_(self, right_r: Any, op: int) -> bool: ...
    def valuation_zero_part(self) -> Any: ...
    def valuation(self) -> int: ...
    def variable(self) -> str: ...
    def prec(self) -> int: ...
    def precision_absolute(self) -> int: ...
    def precision_relative(self) -> int: ...
    def __copy__(self) -> LaurentSeriesRingElement: ...
    def reverse(self, precision: int | None = None) -> LaurentSeriesRingElement: ...
    def derivative(self, *args: Any) -> LaurentSeriesRingElement: ...
    def _derivative(self, var: Any | None = None) -> LaurentSeriesRingElement: ...
    def integral(self) -> LaurentSeriesRingElement: ...
    def nth_root(self, n: int, prec: int | None = None) -> LaurentSeriesRingElement: ...
    def power_series(self) -> Any: ...
    def inverse(self) -> LaurentSeriesRingElement: ...
    def __call__(self, *x: Any, **kwds: Any) -> LaurentSeriesRingElement: ...
    def __pari__(self) -> Any: ...
    def map_coefficients(self, f: Any, new_base_ring: Any = None) -> LaurentSeriesRingElement: ...



# sage.doctest: needs sage.combinat sage.modules
r"""
Kirillov-Reshetikhin tableaux

Kirillov-Reshetikhin tableaux are rectangular tableaux with `r` rows and
`s` columns that naturally arise under the bijection between rigged
configurations and tableaux [RigConBijection]_. They are in bijection with
the elements of the Kirillov-Reshetikhin crystal `B^{r,s}` under the (inverse)
filling map [OSS13]_ [SS2015]_. They do not have to satisfy the semistandard row or column
restrictions. These tensor products are the result from the bijection from
rigged configurations [RigConBijection]_.

For more information, see :class:`~sage.combinat.rigged_configurations.kr_tableaux.KirillovReshetikhinTableaux`
and :class:`~sage.combinat.rigged_configurations.tensor_product_kr_tableaux.TensorProductOfKirillovReshetikhinTableaux`.

REFERENCES:

.. [OSS13] Masato Okado, Reiho Sakamoto, and Anne Schilling.
   *Affine crystal structure on rigged configurations of type* `D_n^{(1)}`.
   J. Algebraic Combinatorics, **37** (2013). 571-599. :arxiv:`1109.3523`.

AUTHORS:

- Travis Scrimshaw (2012-01-03): initial version
- Travis Scrimshaw (2012-11-14): added bijection to KR crystals
"""

# ***************************************************************************
#       Copyright (C) 2012 Travis Scrimshaw <tscrim@ucdavis.edu>
#
#  Distributed under the terms of the GNU General Public License (GPL)
#
#    This code is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#    General Public License for more details.
#
#  The full text of the GPL is available at:
#
#                  https://www.gnu.org/licenses/
# ***************************************************************************

# This contains both the parent and element classes. These should be split if
#   the classes grow larger.

from sage.misc.cachefunc import cached_method
from sage.misc.abstract_method import abstract_method
from sage.misc.lazy_attribute import lazy_attribute
from sage.misc.flatten import flatten

from sage.structure.parent import Parent

from sage.categories.loop_crystals import KirillovReshetikhinCrystals

from sage.combinat.crystals.letters import CrystalOfLetters, EmptyLetter
from sage.combinat.root_system.cartan_type import CartanType
from sage.combinat.crystals.tensor_product import CrystalOfWords
from sage.combinat.crystals.tensor_product import TensorProductOfRegularCrystalsElement
from sage.combinat.crystals.kirillov_reshetikhin import (
    horizontal_dominoes_removed,
    KashiwaraNakashimaTableaux, KirillovReshetikhinGenericCrystalElement,
    partitions_in_box, vertical_dominoes_removed
)
from sage.combinat.partition import Partition
from sage.combinat.tableau import Tableau


class KirillovReshetikhinTableaux(CrystalOfWords):
    r"""
    Kirillov-Reshetikhin tableaux.

    Kirillov-Reshetikhin tableaux are rectangular tableaux with `r` rows and
    `s` columns that naturally arise under the bijection between rigged
    configurations and tableaux [RigConBijection]_. They are in bijection with
    the elements of the Kirillov-Reshetikhin crystal `B^{r,s}` under the
    (inverse) filling map.

    Whenever `B^{r,s} \cong B(s\Lambda_r)` as a classical crystal (which is
    the case for `B^{r,s}` in type `A_n^{(1)}`, `B^{n,s}` in type `C_n^{(1)}`
    and `D_{n+1}^{(2)}`, `B^{n,s}` and `B^{n-1,s}` in type `D_n^{(1)}`) then
    the filling map is trivial.

    For `B^{r,s}` in:

    - type `D_n^{(1)}` when `r \leq n-2`,
    - type `B_n^{(1)}` when `r < n`,
    - type `A_{2n-1}^{(2)}` for all `r`,

    the filling map is defined in [OSS2011]_.

    For the spinor cases in type `D_n^{(1)}`, the crystal `B^{k,s}` where
    `k = n-1, n`,  is isomorphic as a classical crystal to `B(s\Lambda_k)`,
    and here we consider the Kirillov-Reshetikhin tableaux as living in
    `B(2s \Lambda_k)` under the natural doubling map. In this case, the
    crystal operators `e_i` and `f_i` act as `e_i^2` and `f_i^2` respectively.
    See [BijectionDn]_.

    For the spinor case in type `B_n^{(1)}`, the crystal `B^{n,s}`, we
    consider the images under the natural doubling map into `B^{n,2s}`.
    The classical components of this crystal are now given by
    removing `2 \times 2` boxes. The filling map is the same as below
    (see the non-spin type `C_n^{(1)}`).

    For `B^{r,s}` in:

    - type `C_n^{(1)}` when `r < n`,
    - type `A_{2n}^{(2)\dagger}` for all `r`,

    the filling map is given as follows. Suppose we are considering the
    (classically) highest weight element in the classical component
    `B(\lambda)`. Then we fill it in with the horizontal dominoes
    `[\bar{\imath}, i]` in the `i`-th row from the top (in English notation)
    and reordering the columns so that they are increasing. Recall from above
    that `B^{n,s} \cong B(s\Lambda_n)` in type `C^{(1)}_n`.

    For `B^{r,s}` in:

    - type `A_{2n}^{(2)}` for all `r`,
    - type `D_{n+1}^{(2)}` when `r < n`,
    - type `D_4^{(3)}` when `r = 1`,

    the filling map is the same as given in [OSS2011]_ except for
    the rightmost column which is given by the column `[1, 2, \ldots, k,
    \emptyset, \ldots \emptyset]` where `k = (r+x-1)/2` in Step 3 of
    [OSS2011]_.

    For the spinor case in type `D_{n+1}^{(2)}`, the crystal `B^{n,s}`, we
    define the filling map in the same way as in type `D_n^{(1)}`.

    .. NOTE::

        The filling map and classical decompositions in non-spinor cases can
        be classified by how the special node `0` connects with the
        corresponding classical diagram.

    The classical crystal structure is given by the usual Kashiwara-Nakashima
    tableaux rules. That is to embed this into `B(\Lambda_1)^{\otimes n s}`
    by using the reading word and then applying the classical crystal
    operator. The affine crystal structure is given by converting to
    the corresponding KR crystal element, performing the affine crystal
    operator, and pulling back to a KR tableau.

    For more information about the bijection between rigged configurations
    and tensor products of Kirillov-Reshetikhin tableaux, see
    :class:`~sage.combinat.rigged_configurations.tensor_product_kr_tableaux.TensorProductOfKirillovReshetikhinTableaux`.

    .. NOTE::

        The tableaux for all non-simply-laced types are provably correct if the
        bijection with :class:`rigged configurations
        <sage.combinat.rigged_configurations.rigged_configurations.RiggedConfigurations>`
        holds. Therefore this is currently only proven for `B^{r,1}` or
        `B^{1,s}` and in general for types `A_n^{(1)}` and `D_n^{(1)}`.

    INPUT:

    - ``cartan_type`` -- the Cartan type

    - ``r`` -- the Dynkin diagram index (typically the number of rows)

    - ``s`` -- the number of columns

    EXAMPLES::

        sage: KRT = crystals.KirillovReshetikhin(['A', 4, 1], 2, 1, model='KR')
        sage: elt = KRT(4, 3); elt
        [[3], [4]]

        sage: KRT = crystals.KirillovReshetikhin(['D', 4, 1], 2, 1, model='KR')
        sage: elt = KRT(-1, 1); elt
        [[1], [-1]]

    We can create highest weight crystals from a given shape or weight::

        sage: KRT = crystals.KirillovReshetikhin(['D', 4, 1], 2, 2, model='KR')
        sage: KRT.module_generator(shape=[1,1])
        [[1, 1], [2, -1]]
        sage: KRT.module_generator(column_shape=[2])
        [[1, 1], [2, -1]]
        sage: WS = RootSystem(['D',4,1]).weight_space()
        sage: KRT.module_generator(weight=WS.sum_of_terms([[0,-2],[2,1]]))
        [[1, 1], [2, -1]]
        sage: WSC = RootSystem(['D',4]).weight_space()
        sage: KRT.module_generator(classical_weight=WSC.fundamental_weight(2))
        [[1, 1], [2, -1]]

    We can go between
    :func:`~sage.combinat.crystals.kirillov_reshetikhin.KashiwaraNakashimaTableaux`
    and
    :class:`~sage.combinat.rigged_configurations.kr_tableaux.KirillovReshetikhinTableaux`
    elements::

        sage: KRCrys = crystals.KirillovReshetikhin(['D', 4, 1], 2, 2, model='KN')
        sage: KRTab = crystals.KirillovReshetikhin(['D', 4, 1], 2, 2, model='KR')
        sage: elt = KRCrys(3, 2); elt
        [[2], [3]]
        sage: k = KRTab(elt); k
        [[2, 1], [3, -1]]
        sage: KRCrys(k)
        [[2], [3]]

    We check that the classical weights in the classical decompositions
    agree in a few different type::

        sage: KRCrys = crystals.KirillovReshetikhin(['D', 4, 1], 2, 2, model='KN')
        sage: KRTab = crystals.KirillovReshetikhin(['D', 4, 1], 2, 2, model='KR')
        sage: all(t.classical_weight() == KRCrys(t).classical_weight() for t in KRTab)
        True
        sage: KRCrys = crystals.KirillovReshetikhin(['B', 3, 1], 2, 2, model='KN')
        sage: KRTab = crystals.KirillovReshetikhin(['B', 3, 1], 2, 2, model='KR')
        sage: all(t.classical_weight() == KRCrys(t).classical_weight() for t in KRTab)
        True
        sage: KRCrys = crystals.KirillovReshetikhin(['C', 3, 1], 2, 2, model='KN')
        sage: KRTab = crystals.KirillovReshetikhin(['C', 3, 1], 2, 2, model='KR')
        sage: all(t.classical_weight() == KRCrys(t).classical_weight() for t in KRTab)
        True
        sage: KRCrys = crystals.KirillovReshetikhin(['D', 4, 2], 2, 2, model='KN')
        sage: KRTab = crystals.KirillovReshetikhin(['D', 4, 2], 2, 2, model='KR')
        sage: all(t.classical_weight() == KRCrys(t).classical_weight() for t in KRTab)
        True
        sage: KRCrys = crystals.KirillovReshetikhin(['A', 4, 2], 2, 2, model='KN')
        sage: KRTab = crystals.KirillovReshetikhin(['A', 4, 2], 2, 2, model='KR')
        sage: all(t.classical_weight() == KRCrys(t).classical_weight() for t in KRTab)
        True
    """
    @staticmethod
    def __classcall_private__(cls, cartan_type, r, s):
        """
        Normalize the input arguments to ensure unique representation.

        EXAMPLES::

            sage: KRT1 = crystals.KirillovReshetikhin(CartanType(['A',3,1]), 2, 3, model='KR')
            sage: KRT2 = crystals.KirillovReshetikhin(['A',3,1], 2, 3, model='KR')
            sage: KRT1 is KRT2
            True
        """
        ct = CartanType(cartan_type)
        if not ct.is_affine():
            raise ValueError("The Cartan type must be affine")

        typ = ct.type()
        if ct.is_untwisted_affine():
            if typ == 'A':
                return KRTableauxRectangle(ct, r, s)
            if typ == 'B':
                if r == ct.classical().rank():
                    return KRTableauxBn(ct, r, s)
                return KRTableauxTypeVertical(ct, r, s)
            if typ == 'C':
                if r == ct.classical().rank():
                    return KRTableauxRectangle(ct, r, s)
                return KRTableauxTypeHorizonal(ct, r, s)
            if typ == 'D':
                if r == ct.classical().rank() or r == ct.classical().rank() - 1:
                    return KRTableauxSpin(ct, r, s)
                return KRTableauxTypeVertical(ct, r, s)
            if typ == 'E':
                return KRTableauxTypeFromRC(ct, r, s)
        else:
            if typ == 'BC':  # A_{2n}^{(2)}
                return KRTableauxTypeBox(ct, r, s)
            typ = ct.dual().type()
            if typ == 'BC':  # A_{2n}^{(2)\dagger}
                return KRTableauxTypeHorizonal(ct, r, s)
            if typ == 'B':  # A_{2n-1}^{(2)}
                return KRTableauxTypeVertical(ct, r, s)
            if typ == 'C':  # D_{n+1}^{(2)}
                if r == ct.dual().classical().rank():
                    return KRTableauxDTwistedSpin(ct, r, s)
                return KRTableauxTypeBox(ct, r, s)
            # if typ == 'F':  # E_6^{(2)}
            if typ == 'G':  # D_4^{(3)}
                if r == 1:
                    return KRTableauxTypeBox(ct, r, s)
                return KRTableauxTypeFromRC(ct, r, s)

        raise NotImplementedError
        # return super(KirillovReshetikhinTableaux, cls).__classcall__(cls, ct, r, s)

    def __init__(self, cartan_type, r, s):
        r"""
        Initialize ``self``.

        EXAMPLES::

            sage: KRT = crystals.KirillovReshetikhin(['A', 4, 1], 2, 2, model='KR')
            sage: TestSuite(KRT).run()
            sage: KRT = crystals.KirillovReshetikhin(['D', 4, 1], 2, 2, model='KR')
            sage: TestSuite(KRT).run()  # long time
            sage: KRT = crystals.KirillovReshetikhin(['D', 4, 1], 4, 1, model='KR'); KRT
            Kirillov-Reshetikhin tableaux of type ['D', 4, 1] and shape (4, 1)
            sage: TestSuite(KRT).run()
        """
        self._r = r
        self._s = s
        self._cartan_type = cartan_type

        Parent.__init__(self, category=KirillovReshetikhinCrystals())

        self.letters = CrystalOfLetters(cartan_type.classical())
        self.module_generators = self._build_module_generators()

    def _repr_(self):
        """
        Return a string representation of ``self``.

        EXAMPLES::

            sage: crystals.KirillovReshetikhin(['A', 4, 1], 2, 3, model='KR')
            Kirillov-Reshetikhin tableaux of type ['A', 4, 1] and shape (2, 3)
        """
        return "Kirillov-Reshetikhin tableaux of type {} and shape ({}, {})".format(
            self._cartan_type, self._r, self._s)

    def __iter__(self):
        """
        Return the iterator of ``self``.

        EXAMPLES::

            sage: KR = crystals.KirillovReshetikhin(['A', 5, 2], 2, 1, model='KR')
            sage: L = [x for x in KR]
            sage: len(L)
            15
        """
        index_set = self._cartan_type.classical().index_set()
        from sage.sets.recursively_enumerated_set import RecursivelyEnumeratedSet
        rset = RecursivelyEnumeratedSet(self.module_generators,
                                        lambda x: [x.f(i) for i in index_set],
                                        structure='graded')
        return rset.breadth_first_search_iterator()

    def module_generator(self, i=None, **options):
        r"""
        Return the specified module generator.

        INPUT:

        - ``i`` -- the index of the module generator

        We can also get a module generator by using one of the following
        optional arguments:

        - ``shape`` -- the associated shape
        - ``column_shape`` -- the shape given as columns (a column of length
          `k` correspond to a classical weight `\omega_k`)
        - ``weight`` -- the weight
        - ``classical_weight`` -- the classical weight

        If no arguments are specified, then return the unique module generator
        of classical weight `s \Lambda_r`.

        EXAMPLES::

            sage: KRT = crystals.KirillovReshetikhin(['D', 4, 1], 2, 2, model='KR')
            sage: KRT.module_generator(1)
            [[1, 1], [2, -1]]
            sage: KRT.module_generator(shape=[1,1])
            [[1, 1], [2, -1]]
            sage: KRT.module_generator(column_shape=[2])
            [[1, 1], [2, -1]]
            sage: WS = RootSystem(['D',4,1]).weight_space()
            sage: KRT.module_generator(weight=WS.sum_of_terms([[0,-2],[2,1]]))
            [[1, 1], [2, -1]]
            sage: WSC = RootSystem(['D',4]).weight_space()
            sage: KRT.module_generator(classical_weight=WSC.fundamental_weight(2))
            [[1, 1], [2, -1]]
            sage: KRT.module_generator()
            [[1, 1], [2, 2]]

            sage: KRT = crystals.KirillovReshetikhin(['A', 3, 1], 2, 2, model='KR')
            sage: KRT.module_generator()
            [[1, 1], [2, 2]]
        """
        if i is not None:
            return self.module_generators[i]
        n = self._cartan_type.classical().rank()

        if "shape" in options:
            shape = list(options["shape"])
            # Make sure the shape is the correct length
            if len(shape) < n:
                shape.extend([0] * (n - len(shape)))
            for mg in self.module_generators:
                if list(mg.classical_weight().to_vector()) == shape:
                    return mg
            return None

        if "column_shape" in options:
            shape = list(Partition(options["column_shape"]).conjugate())
            if len(shape) < n:
                shape.extend([0] * (n - len(shape)))
            for mg in self.module_generators:
                if list(mg.classical_weight().to_vector()) == shape:
                    return mg
            return None

        if "weight" in options:
            wt = options["weight"]
            for mg in self.module_generators:
                if mg.weight() == wt:
                    return mg
            return None

        if "classical_weight" in options:
            wt = options["classical_weight"]
            for mg in self.module_generators:
                if mg.classical_weight() == wt:
                    return mg
            return None

        # Otherwise return the unique module generator of classical weight `s \Lambda_r`
        R = self.weight_lattice_realization()
        Lambda = R.fundamental_weights()
        r = self.r()
        s = self.s()
        weight = s * Lambda[r] - s * Lambda[0] * Lambda[r].level() / Lambda[0].level()
        for b in self.module_generators:
            if b.weight() == weight:
                return b
        assert False

    @abstract_method
    def _build_module_generators(self):
        """
        Build the module generators.

        EXAMPLES::

            sage: KRT = crystals.KirillovReshetikhin(['A', 4, 1], 2, 3, model='KR')
            sage: KRT._build_module_generators()
            ([[1, 1, 1], [2, 2, 2]],)
        """

    @abstract_method(optional=True)
    def from_kirillov_reshetikhin_crystal(self, krc):
        """
        Construct an element of ``self`` from the Kirillov-Reshetikhin
        crystal element ``krc``.

        EXAMPLES::

            sage: KRT = crystals.KirillovReshetikhin(['A', 4, 1], 2, 1, model='KR')
            sage: C = crystals.KirillovReshetikhin(['A',4,1], 2, 1, model='KN')
            sage: krc = C(4,3); krc
            [[3], [4]]
            sage: KRT.from_kirillov_reshetikhin_crystal(krc)
            [[3], [4]]
        """

    def _element_constructor_(self, *lst, **options):
        """
        Construct a
        :class:`~sage.combinat.rigged_configurations.kr_tableaux.KirillovReshetikhinTableauxElement`.

        EXAMPLES::

            sage: KRT = crystals.KirillovReshetikhin(['A', 4, 1], 2, 1, model='KR')
            sage: KRT(3, 4) # indirect doctest
            [[4], [3]]
            sage: KRT(4, 3)
            [[3], [4]]
        """
        if isinstance(lst[0], KirillovReshetikhinGenericCrystalElement):
            # Check to make sure it can be converted
            if lst[0].cartan_type() != self.cartan_type() \
               or lst[0].parent().r() != self._r or lst[0].parent().s() != self._s:
                raise ValueError("the Kirillov-Reshetikhin crystal must have the same Cartan type and (r,s)")
            return self.from_kirillov_reshetikhin_crystal(lst[0])

        return self.element_class(self, list(lst), **options)

    def r(self):
        """
        Return the value `r` for this tableaux class which corresponds to the
        number of rows.

        EXAMPLES::

            sage: KRT = crystals.KirillovReshetikhin(['A', 4, 1], 2, 1, model='KR')
            sage: KRT.r()
            2
        """
        return self._r

    def s(self):
        """
        Return the value `s` for this tableaux class which corresponds to the
        number of columns.

        EXAMPLES::

            sage: KRT = crystals.KirillovReshetikhin(['A', 4, 1], 2, 1, model='KR')
            sage: KRT.s()
            1
        """
        return self._s

    @cached_method
    def kirillov_reshetikhin_crystal(self):
        """
        Return the corresponding KR crystal in the
        :func:`Kashiwara-Nakashima model
        <sage.combinat.crystals.kirillov_reshetikhin.KashiwaraNakashimaTableaux>`.

        EXAMPLES::

            sage: crystals.KirillovReshetikhin(['A', 4, 1], 2, 1, model='KR').kirillov_reshetikhin_crystal()
            Kirillov-Reshetikhin crystal of type ['A', 4, 1] with (r,s)=(2,1)
        """
        return KashiwaraNakashimaTableaux(self._cartan_type, self._r, self._s)

    def classical_decomposition(self):
        """
        Return the classical crystal decomposition of ``self``.

        EXAMPLES::

            sage: crystals.KirillovReshetikhin(['D', 4, 1], 2, 2, model='KR').classical_decomposition()
            The crystal of tableaux of type ['D', 4] and shape(s) [[], [1, 1], [2, 2]]
        """
        return self.kirillov_reshetikhin_crystal().classical_decomposition()

    def tensor(self, *crystals, **options):
        """
        Return the tensor product of ``self`` with ``crystals``.

        If ``crystals`` is a list of (a tensor product of) KR tableaux, this
        returns a
        :class:`~sage.combinat.rigged_configurations.tensor_product_kr_tableaux.TensorProductOfKirillovReshetikhinTableaux`.

        EXAMPLES::

            sage: K = crystals.KirillovReshetikhin(['A', 3, 1], 2, 2, model='KR')
            sage: TP = crystals.TensorProductOfKirillovReshetikhinTableaux(['A', 3, 1], [[1,3],[3,1]])
            sage: K.tensor(TP, K)
            Tensor product of Kirillov-Reshetikhin tableaux of type ['A', 3, 1]
             and factor(s) ((2, 2), (1, 3), (3, 1), (2, 2))

            sage: C = crystals.KirillovReshetikhin(['A',3,1], 3, 1, model='KN')
            sage: K.tensor(K, C)
            Full tensor product of the crystals
             [Kirillov-Reshetikhin tableaux of type ['A', 3, 1] and shape (2, 2),
              Kirillov-Reshetikhin tableaux of type ['A', 3, 1] and shape (2, 2),
              Kirillov-Reshetikhin crystal of type ['A', 3, 1] with (r,s)=(3,1)]
        """
        ct = self._cartan_type
        from sage.combinat.rigged_configurations.tensor_product_kr_tableaux \
            import TensorProductOfKirillovReshetikhinTableaux
        if all(isinstance(B, (KirillovReshetikhinTableaux, TensorProductOfKirillovReshetikhinTableaux))
               and B.cartan_type() == ct for B in crystals):
            dims = [[self._r, self._s]]
            for B in crystals:
                if isinstance(B, TensorProductOfKirillovReshetikhinTableaux):
                    dims += B.dims
                elif isinstance(B, KirillovReshetikhinTableaux):
                    dims.append([B._r, B._s])
            return TensorProductOfKirillovReshetikhinTableaux(ct, dims)
        return super().tensor(*crystals, **options)

    @lazy_attribute
    def _tableau_height(self):
        """
        The height of the tableaux in ``self``.

        EXAMPLES::

            sage: K = crystals.KirillovReshetikhin(['A', 3, 1], 3, 2, model='KR')
            sage: K._tableau_height
            3
        """
        return self._r


class KRTableauxRectangle(KirillovReshetikhinTableaux):
    r"""
    Kirillov-Reshetkhin tableaux `B^{r,s}` whose module generator is a single
    `r \times s` rectangle.

    These are Kirillov-Reshetkhin tableaux `B^{r,s}` of type:

    - `A_n^{(1)}` for all `1 \leq r \leq n`,
    - `C_n^{(1)}` when `r = n`.

    TESTS::

        sage: KRT = crystals.KirillovReshetikhin(['A', 3, 1], 2, 2, model='KR')
        sage: TestSuite(KRT).run()
        sage: KRT = crystals.KirillovReshetikhin(['C', 3, 1], 3, 2, model='KR')
        sage: TestSuite(KRT).run() # long time
    """

    def _build_module_generators(self):
        r"""
        Build the module generators.

        There is only one module generator which corresponds to a single
        `r \times s` rectangle.

        EXAMPLES::

            sage: KRT = crystals.KirillovReshetikhin(['A', 4, 1], 2, 3, model='KR')
            sage: KRT._build_module_generators()
            ([[1, 1, 1], [2, 2, 2]],)
        """
        tableau = [[self._r - j for j in range(self._r)]
                   for i in range(self._s)]

        return (self.element_class(self, [self.letters(x) for x in flatten(tableau)]),)

    def from_kirillov_reshetikhin_crystal(self, krc):
        """
        Construct a
        :class:`~sage.combinat.rigged_configurations.kr_tableaux.KirillovReshetikhinTableauxElement`.

        EXAMPLES::

            sage: KRT = crystals.KirillovReshetikhin(['A', 4, 1], 2, 1, model='KR')
            sage: C = crystals.KirillovReshetikhin(['A',4,1], 2, 1, model='KN')
            sage: krc = C(4,3); krc
            [[3], [4]]
            sage: KRT.from_kirillov_reshetikhin_crystal(krc)
            [[3], [4]]
        """
        # To build a KR tableau from a KR crystal:
        # 1 - start with the highest weight KR tableau
        # 2 - determine a path from the KR crystal to its highest weight
        # 3 - apply the inverse path to the highest weight KR tableau
        f_str = reversed(krc.lift().to_highest_weight()[1])
        return self.module_generators[0].f_string(f_str)


class KRTableauxTypeVertical(KirillovReshetikhinTableaux):
    r"""
    Kirillov-Reshetkihn tableaux `B^{r,s}` of type:

    - `D_n^{(1)}` for all `1 \leq r < n-1`,
    - `B_n^{(1)}` for all `1 \leq r < n`,
    - `A_{2n-1}^{(2)}` for all `1 \leq r \leq n`.

    TESTS::

        sage: KRT = crystals.KirillovReshetikhin(['D', 4, 1], 1, 1, model='KR')
        sage: TestSuite(KRT).run()
        sage: KRT = crystals.KirillovReshetikhin(['B', 3, 1], 2, 2, model='KR')
        sage: TestSuite(KRT).run() # long time
        sage: KRT = crystals.KirillovReshetikhin(['A', 5, 2], 2, 2, model='KR')
        sage: TestSuite(KRT).run() # long time
    """

    def _fill(self, weight):
        r"""
        Return the highest weight KR tableau of weight ``weight``.

        INPUT:

        - ``weight`` -- the weight of the highest weight KR tableau (the
          conjugate of the shape of the KR crystal's tableau)

        OUTPUT: a `r \times s` tableau

        EXAMPLES::

            sage: KRT = crystals.KirillovReshetikhin(['D', 4, 1], 2, 1, model='KR')
            sage: KRT._fill([])
            [[1], [-1]]
            sage: KRT = crystals.KirillovReshetikhin(['D', 14, 1], 12, 7, model='KR')
            sage: KRT._fill([10,10,8,2,2,2])
            [[1, 1, 1, 1, 1, 7, 1], [2, 2, 2, 2, 2, 8, 2], [3, 3, 7, 9, 7, 9, 3], [4, 4, 8, 10, 8, 10, 4], [5, 5, 9, 11, 9, 11, 5], [6, 6, 10, 12, 10, 12, 6], [7, 7, 11, -12, 11, -12, 7], [8, 8, 12, -11, 12, -11, 8], [9, 9, -12, -10, -12, -10, 9], [10, 10, -11, -9, -11, -9, -9], [-12, 11, -10, -8, -10, -8, -8], [-11, 12, -9, -7, -9, -7, -7]]
            sage: KRT._fill([10,10,6,2,2,2])
            [[1, 1, 1, 1, 1, 5, 1], [2, 2, 2, 2, 2, 6, 2], [3, 3, 9, 7, 9, 7, 3], [4, 4, 10, 8, 10, 8, 4], [5, 5, 11, 9, 11, 9, 5], [6, 6, 12, 10, 12, 10, 6], [7, 7, -12, 11, -12, 11, 7], [8, 8, -11, 12, -11, 12, 8], [9, 9, -10, -12, -10, -12, -8], [10, 10, -9, -11, -9, -11, -7], [-12, 11, -8, -10, -8, -10, -6], [-11, 12, -7, -9, -7, -9, -5]]
        """
        # Add zeros until the shape has length s
        weight_list = list(weight)  # Make sure we have a list
        while len(weight_list) != self._s:
            weight_list.append(0)

        tableau = []
        i = 0
        # Step 0 - Fill first columns of height r
        while i < self._s and weight_list[i] == self._r:
            tableau.append([self._r - j for j in range(self._r)])
            i += 1

        # Step 1 - Add the alternating columns until we hit an odd number of columns
        c = -1
        while i < self._s:
            # If it is an odd number of columns
            if i == self._s - 1 or weight_list[i] != weight_list[i + 1]:
                c = weight_list[i]
                i += 1
                break
            temp_list = [-(weight_list[i] + j + 1) for j in range(self._r - weight_list[i])]
            for j in range(weight_list[i]):
                temp_list.append(weight_list[i] - j)
            tableau.append(temp_list)
            tableau.append([self._r - j for j in range(self._r)])
            i += 2

        # Step 2 - Add the x dependent columns
        x = c + 1
        while i < self._s:
            temp_list = [-x - j for j in range(self._r - x + 1)]  # +1 for indexing
            for j in range(x - weight_list[i] - 1):  # +1 for indexing
                temp_list.append(self._r - j)
            x = temp_list[-1]  # This is the h+1 entry of the column
            for j in range(weight_list[i]):
                temp_list.append(weight_list[i] - j)

            tableau.append(temp_list)
            i += 1

        # Step 3 - Add the final column
        if c > -1:
            val = (self._r + x - 1) // 2
            temp_list = [-x - j for j in range(self._r - val)]
            for j in range(val):
                temp_list.append(val - j)
            tableau.append(temp_list)

        return self.element_class(self, [self.letters(x) for x in flatten(tableau)])

    def _build_module_generators(self):
        """
        Build the module generators.

        EXAMPLES::

            sage: KRT = crystals.KirillovReshetikhin(['D',4,1], 2, 3, model='KR')
            sage: KRT._build_module_generators()
            ([[-2, 1, 1], [-1, 2, -1]], [[1, -2, 1], [2, -1, 2]],
             [[1, 1, 1], [2, 2, -1]], [[1, 1, 1], [2, 2, 2]])
        """
        return tuple(self._fill(weight) for weight in
                     horizontal_dominoes_removed(self._s, self._r))

    def from_kirillov_reshetikhin_crystal(self, krc):
        """
        Construct an element of ``self`` from the Kirillov-Reshetikhin
        crystal element ``krc``.

        EXAMPLES::

            sage: KRT = crystals.KirillovReshetikhin(['D',4,1], 2, 3, model='KR')
            sage: C = crystals.KirillovReshetikhin(['D',4,1], 2, 3, model='KN')
            sage: krc = C(4,3); krc
            [[3], [4]]
            sage: KRT.from_kirillov_reshetikhin_crystal(krc)
            [[3, -2, 1], [4, -1, 2]]
        """
        # To build a KR tableau from a KR crystal:
        # 1 - start with a highest weight KR tableau generated from the
        #  shape of the KR crystal
        # 2 - determine a path from the KR crystal to its highest weight
        # 3 - apply the inverse path to the highest weight KR tableau
        lifted = krc.lift()
        weight = lifted.to_tableau().shape().conjugate()
        f_str = reversed(lifted.to_highest_weight()[1])
        return self._fill(weight).f_string(f_str)


class KRTableauxTypeHorizonal(KirillovReshetikhinTableaux):
    r"""
    Kirillov-Reshetikhin tableaux `B^{r,s}` of type:

    - `C_n^{(1)}` for `1 \leq r < n`,
    - `A_{2n}^{(2)\dagger}` for `1 \leq r \leq n`.

    TESTS::

        sage: KRT = crystals.KirillovReshetikhin(['C', 3, 1], 2, 2, model='KR')
        sage: TestSuite(KRT).run() # long time
        sage: KRT = crystals.KirillovReshetikhin(CartanType(['A', 4, 2]).dual(), 2, 2, model='KR')
        sage: TestSuite(KRT).run()
    """

    def _fill(self, shape):
        r"""
        Return the highest weight KR tableau of weight ``shape``.

        INPUT:

        - ``shape`` -- the shape of the KR crystal's tableau

        OUTPUT: a `r \times s` tableau

        EXAMPLES::

            sage: KRT = crystals.KirillovReshetikhin(['C', 5, 1], 3, 5, model='KR')
            sage: KRT._fill([3,3,1])
            [[1, 1, 1, -3, 1], [2, 2, 2, -2, 2], [3, -3, 3, -1, 3]]
            sage: KRT = crystals.KirillovReshetikhin(['C', 10, 1], 5, 6, model='KR')
            sage: KRT._fill([6,4,2,2])
            [[1, 1, 1, 1, 1, 1], [2, 2, 2, 2, -5, 2], [3, 3, -5, 3, -4, 3], [4, 4, -4, 4, -3, 4], [-5, 5, -3, 5, -2, 5]]
            sage: KRT._fill([6,4])
            [[1, 1, 1, 1, 1, 1], [2, 2, 2, 2, -5, 2], [-5, 3, -5, 3, -4, 3], [-4, 4, -4, 4, -3, 4], [-3, 5, -3, 5, -2, 5]]
        """
        # Add zeros until the shape has length s
        shape_list = list(shape)  # Make sure we have a list
        while len(shape_list) != self._r:
            shape_list.append(0)

        lst = []
        for col in range(1, self._s + 1):
            if (self._s - col) % 2 == 0:
                lst.extend(self.letters(self._r - x) for x in range(self._r))
            else:
                m = self._r
                for j, val in enumerate(shape_list):
                    if col >= val:
                        m = j
                        break
                lst.extend(self.letters(-x) for x in range(m + 1, self._r + 1))
                lst.extend(self.letters(m - x) for x in range(m))

        return self.element_class(self, lst)

    def _build_module_generators(self):
        """
        Build the module generators.

        EXAMPLES::

            sage: KRT = crystals.KirillovReshetikhin(['C',4,1], 2, 3, model='KR')
            sage: KRT._build_module_generators()
            ([[1, -2, 1], [2, -1, 2]], [[1, 1, 1], [2, -2, 2]], [[1, 1, 1], [2, 2, 2]])
        """
        return tuple(self._fill(shape) for shape in horizontal_dominoes_removed(self._r, self._s))

    def from_kirillov_reshetikhin_crystal(self, krc):
        """
        Construct an element of ``self`` from the Kirillov-Reshetikhin
        crystal element ``krc``.

        EXAMPLES::

            sage: KRT = crystals.KirillovReshetikhin(['C',4,1], 2, 3, model='KR')
            sage: C = crystals.KirillovReshetikhin(['C',4,1], 2, 3, model='KN')
            sage: krc = C(4,3); krc
            [[3], [4]]
            sage: KRT.from_kirillov_reshetikhin_crystal(krc)
            [[3, -2, 1], [4, -1, 2]]
        """
        # To build a KR tableau from a KR crystal:
        # 1 - start with a highest weight KR tableau generated from the
        #  shape of the KR crystal
        # 2 - determine a path from the KR crystal to its highest weight
        # 3 - apply the inverse path to the highest weight KR tableau
        lifted = krc.lift()
        shape = lifted.to_tableau().shape()
        f_str = reversed(lifted.to_highest_weight()[1])
        return self._fill(shape).f_string(f_str)


class KRTableauxTypeBox(KRTableauxTypeVertical):
    r"""
    Kirillov-Reshetikhin tableaux `B^{r,s}` of type:

    - `A_{2n}^{(2)}` for all `r \leq n`,
    - `D_{n+1}^{(2)}` for all `r < n`,
    - `D_4^{(3)}` for `r = 1`.

    TESTS::

        sage: KRT = crystals.KirillovReshetikhin(['A', 4, 2], 2, 2, model='KR')
        sage: TestSuite(KRT).run()
        sage: KRT = crystals.KirillovReshetikhin(['D', 4, 2], 2, 2, model='KR')
        sage: TestSuite(KRT).run() # long time
        sage: KRT = crystals.KirillovReshetikhin(['D', 4, 3], 1, 2, model='KR')
        sage: TestSuite(KRT).run() # long time
    """

    def _fill(self, weight):
        r"""
        Return the highest weight KR tableau of weight ``weight``.

        INPUT:

        - ``weight`` -- the weight of the highest weight KR tableau (the
          conjugate of the shape of the KR crystal's tableau)

        OUTPUT: a `r \times s` tableau

        EXAMPLES::

            sage: KRT = crystals.KirillovReshetikhin(['D', 4, 1], 2, 1, model='KR')
            sage: KRT._fill([])
            [[1], [-1]]
            sage: KRT = crystals.KirillovReshetikhin(['D', 14, 1], 12, 7, model='KR')
            sage: KRT._fill([10,10,8,2,2,2])
            [[1, 1, 1, 1, 1, 7, 1], [2, 2, 2, 2, 2, 8, 2], [3, 3, 7, 9, 7, 9, 3], [4, 4, 8, 10, 8, 10, 4], [5, 5, 9, 11, 9, 11, 5], [6, 6, 10, 12, 10, 12, 6], [7, 7, 11, -12, 11, -12, 7], [8, 8, 12, -11, 12, -11, 8], [9, 9, -12, -10, -12, -10, 9], [10, 10, -11, -9, -11, -9, -9], [-12, 11, -10, -8, -10, -8, -8], [-11, 12, -9, -7, -9, -7, -7]]
            sage: KRT._fill([10,10,6,2,2,2])
            [[1, 1, 1, 1, 1, 5, 1], [2, 2, 2, 2, 2, 6, 2], [3, 3, 9, 7, 9, 7, 3], [4, 4, 10, 8, 10, 8, 4], [5, 5, 11, 9, 11, 9, 5], [6, 6, 12, 10, 12, 10, 6], [7, 7, -12, 11, -12, 11, 7], [8, 8, -11, 12, -11, 12, 8], [9, 9, -10, -12, -10, -12, -8], [10, 10, -9, -11, -9, -11, -7], [-12, 11, -8, -10, -8, -10, -6], [-11, 12, -7, -9, -7, -9, -5]]
        """
        # Add zeros until the shape has length s
        weight_list = list(weight)  # Make sure we have a list
        while len(weight_list) != self._s:
            weight_list.append(0)

        tableau = []
        i = 0
        # Step 0 - Fill first columns of height r
        while i < self._s and weight_list[i] == self._r:
            tableau.append([self._r - j for j in range(self._r)])
            i += 1

        # Step 1 - Add the alternating columns until we hit an odd number of columns
        c = -1
        while i < self._s:
            # If it is an odd number of columns
            if i == self._s - 1 or weight_list[i] != weight_list[i + 1]:
                c = weight_list[i]
                i += 1
                break
            temp_list = [-(weight_list[i] + j + 1) for j in range(self._r - weight_list[i])]
            for j in range(weight_list[i]):
                temp_list.append(weight_list[i] - j)
            tableau.append(temp_list)
            tableau.append([self._r - j for j in range(self._r)])
            i += 2

        # Step 2 - Add the x dependent columns
        x = c + 1
        while i < self._s:
            temp_list = [-x - j for j in range(self._r - x + 1)]  # +1 for indexing
            for j in range(x - weight_list[i] - 1):  # +1 for indexing
                temp_list.append(self._r - j)
            x = temp_list[-1]  # This is the h+1 entry of the column
            for j in range(weight_list[i]):
                temp_list.append(weight_list[i] - j)

            tableau.append(temp_list)
            i += 1

        # Step 3 - Add the final column
        if c > -1:
            val = x - 1
            temp_list = ['E' for j in range(self._r - val)]
            for j in range(val):
                temp_list.append(val - j)
            tableau.append(temp_list)

        return self.element_class(self, [self.letters(x) for x in flatten(tableau)])

    def _build_module_generators(self):
        """
        Build the module generators.

        EXAMPLES::

            sage: KRT = crystals.KirillovReshetikhin(['A',4,2], 2, 2, model='KR')
            sage: KRT._build_module_generators()
            ([[-2, 1], [-1, 2]], [[2, 1], [-2, E]], [[1, E], [2, E]],
             [[1, 1], [-2, 2]], [[1, 1], [2, E]], [[1, 1], [2, 2]])
        """
        return tuple(self._fill(weight) for weight in partitions_in_box(self._s, self._r))


class KRTableauxSpin(KRTableauxRectangle):
    r"""
    Kirillov-Reshetikhin tableaux `B^{r,s}` of type `D_n^{(1)}` with
    `r = n, n-1`.

    TESTS::

        sage: KRT = crystals.KirillovReshetikhin(['D', 4, 1], 3, 2, model='KR')
        sage: TestSuite(KRT).run()
        sage: KRT = crystals.KirillovReshetikhin(['D', 4, 1], 4, 2, model='KR')
        sage: TestSuite(KRT).run()
    """

    def _build_module_generators(self):
        r"""
        Build the module generators.

        There is only one module generator which corresponds to a single
        `n \times s` rectangle.

        EXAMPLES::

            sage: KRT = crystals.KirillovReshetikhin(['D', 4, 1], 3, 3, model='KR')
            sage: KRT._build_module_generators()
            ([[1, 1, 1], [2, 2, 2], [3, 3, 3], [-4, -4, -4]],)
            sage: KRT = crystals.KirillovReshetikhin(['D', 4, 1], 4, 3, model='KR')
            sage: KRT._build_module_generators()
            ([[1, 1, 1], [2, 2, 2], [3, 3, 3], [4, 4, 4]],)
        """
        n = self.cartan_type().classical().rank()
        if self._r == n:
            return KRTableauxRectangle._build_module_generators(self)

        tableau = []
        for i in range(self._s):
            tableau.append([-n] + [self._r - j for j in range(self._r)])

        return (self.element_class(self, [self.letters(x) for x in flatten(tableau)]),)


class KRTableauxBn(KRTableauxTypeHorizonal):
    """
    Kirillov-Reshetkhin tableaux `B^{n,s}` of type `B_n^{(1)}`.

    TESTS::

        sage: KRT = crystals.KirillovReshetikhin(['B', 2, 1], 2, 3, model='KR')
        sage: TestSuite(KRT).run()
    """

    def _build_module_generators(self):
        """
        Build the module generators.

        EXAMPLES::

            sage: KRT = crystals.KirillovReshetikhin(['B', 2, 1], 2, 2, model='KR')
            sage: KRT._build_module_generators()
            ([[-2, 1], [-1, 2]], [[1, 1], [2, 2]])
        """
        odd = int(self._s % 2)
        shapes = ([int(x * 2 + odd) for x in sh]
                  for sh in vertical_dominoes_removed(self._r, self._s // 2))
        return tuple(self._fill(sh) for sh in shapes)

    def from_kirillov_reshetikhin_crystal(self, krc):
        """
        Construct an element of ``self`` from the Kirillov-Reshetikhin
        crystal element ``krc``.

        EXAMPLES::

            sage: KR = crystals.KirillovReshetikhin(['B',3,1], 3, 3, model='KR')
            sage: C = crystals.KirillovReshetikhin(['B',3,1], 3, 3, model='KN')
            sage: krc = C.module_generators[1].f_string([3,2,3,1,3,3]); krc
            [++-, [[2], [0], [-3]]]
            sage: KR.from_kirillov_reshetikhin_crystal(krc)
            [[1, 1, 2], [2, 2, -3], [-3, -3, -1]]
        """
        # To build a KR tableau from a type B_n spinor KR crystal:
        # 1 - determine a path from the KR crystal to its highest weight
        # 2 - find the corresponding highest weight KR tableau
        # 3 - apply the inverse path to the highest weight KR tableau
        lifted = krc.lift()
        to_hw = lifted.to_highest_weight()
        f_str = reversed(to_hw[1])
        wt = to_hw[0].weight()
        for x in self.module_generators:
            if x.classical_weight() == wt:
                return x.f_string(f_str)
        raise ValueError("no matching highest weight element found")


class KirillovReshetikhinTableauxElement(TensorProductOfRegularCrystalsElement):
    r"""
    A Kirillov-Reshetikhin tableau.

    For more information, see
    :class:`~sage.combinat.rigged_configurations.kr_tableaux.KirillovReshetikhinTableaux`
    and
    :class:`~sage.combinat.rigged_configurations.tensor_product_kr_tableaux.TensorProductOfKirillovReshetikhinTableaux`.
    """

    def __init__(self, parent, list, **options):
        r"""
        Initialize ``self``.

        EXAMPLES::

            sage: KRT = crystals.KirillovReshetikhin(['A', 4, 1], 2, 1, model='KR')
            sage: elt = KRT(4, 3); elt
            [[3], [4]]
            sage: TestSuite(elt).run()
        """
        # Make sure we are a list of letters
        if list and not isinstance(list[0], (parent.letters.element_class, EmptyLetter)):
            list = [parent.letters(x) for x in list]
        TensorProductOfRegularCrystalsElement.__init__(self, parent, list)

    def _repr_(self):
        """
        Return the string representation of ``self``.

        EXAMPLES::

            sage: KRT = crystals.KirillovReshetikhin(['A', 4, 1], 2, 1, model='KR')
            sage: KRT(3,2)
            [[2], [3]]
        """
        return repr(self.to_array())

    def _repr_diagram(self):
        """
        Return a string representation of ``self`` as a diagram.

        EXAMPLES::

            sage: KRT = crystals.KirillovReshetikhin(['A',4,1], 2, 2, model='KR')
            sage: elt = KRT(2,1,4,3)
            sage: print(elt._repr_diagram())
              1  3
              2  4
        """
        return self.to_tableau()._repr_diagram()

    def _latex_(self):
        r"""
        Return a latex representation of ``self``.

        EXAMPLES::

            sage: KRT = crystals.KirillovReshetikhin(['A', 4, 1], 2, 3, model='KR')
            sage: latex(KRT(3,2,4,2,4,3))
            {\def\lr#1{\multicolumn{1}{|@{\hspace{.6ex}}c@{\hspace{.6ex}}|}{\raisebox{-.3ex}{$#1$}}}
            \raisebox{-.6ex}{$\begin{array}[b]{*{3}c}\cline{1-3}
            \lr{2}&\lr{2}&\lr{3}\\\cline{1-3}
            \lr{3}&\lr{4}&\lr{4}\\\cline{1-3}
            \end{array}$}
            }
        """
        from sage.combinat.output import tex_from_array
        return tex_from_array([[val._latex_() for val in row] for row in self.to_array()])

    def _ascii_art_(self):
        r"""
        Return an ASCII art representation of ``self``.

        EXAMPLES::

            sage: KRT = crystals.KirillovReshetikhin(['A',4,1], 2, 2, model='KR')
            sage: ascii_art(KRT(2,1,4,3))
              1  3
              2  4
        """
        from sage.typeset.ascii_art import AsciiArt
        return AsciiArt(self._repr_diagram().splitlines())

    def _unicode_art_(self):
        r"""
        Return a unicode art representation of ``self``.

        EXAMPLES::

            sage: KRT = crystals.KirillovReshetikhin(['D',4,1], 2, 2, model='KR')
            sage: unicode_art(KRT(2,1,-4,3))
            ┌───┬───┐
            │ 1 │ 3 │
            ├───┼───┤
            │ 2 │ 4̄ │
            └───┴───┘
        """
        return self.to_tableau()._unicode_art_()

    def to_kirillov_reshetikhin_crystal(self):
        r"""
        Construct a
        :func:`~sage.combinat.crystals.kirillov_reshetihkin.KashiwaraNakashimaTableaux`
        element from ``self``.

        We construct the Kirillov-Reshetikhin crystal element as follows:

        1. Determine the shape `\lambda` of the KR crystal from the weight.
        2. Determine a path `e_{i_1} e_{i_2} \cdots e_{i_k}` to the highest
           weight.
        3. Apply `f_{i_k} \cdots f_{i_2} f_{i_1}` to a highest weight KR
           crystal of shape `\lambda`.

        EXAMPLES::

            sage: KRT = crystals.KirillovReshetikhin(['D',4,1], 2, 2, model='KR')
            sage: elt = KRT(3,2,-1,1); elt
            [[2, 1], [3, -1]]
            sage: elt.to_kirillov_reshetikhin_crystal()
            [[2], [3]]

        TESTS:

        Spinor tests::

            sage: KRT = crystals.KirillovReshetikhin(['D',4,1], 4, 3, model='KR')
            sage: KRC = crystals.KirillovReshetikhin(['D',4,1], 4, 3, model='KN')
            sage: elt = KRT(-3,-4,2,1,-3,-4,2,1,-2,-4,3,1); elt
            [[1, 1, 1], [2, 2, 3], [-4, -4, -4], [-3, -3, -2]]
            sage: ret = elt.to_kirillov_reshetikhin_crystal(); ret
            [++--, [[1], [3], [-4], [-3]]]
            sage: test = KRT(ret); test
            [[1, 1, 1], [2, 2, 3], [-4, -4, -4], [-3, -3, -2]]
            sage: test == elt
            True
        """
        return self.parent().kirillov_reshetikhin_crystal()(self)

    @cached_method
    def to_array(self, rows=True):
        r"""
        Return a 2-dimensional array representation of this
        Kirillov-Reshetikhin element.

        If the output is in rows, then it outputs the top row first (in the
        English convention) from left to right.

        For example: if the reading word is `[2, 1, 4, 3]`, so as a
        `2 \times 2` tableau::

            1 3
            2 4

        we output ``[[1, 3], [2, 4]]``.

        If the output is in columns, then it outputs the leftmost column first
        with the bottom element first. In other words this parses the reading
        word into its columns.

        Continuing with the previous example, the output would be
        ``[[2, 1], [4, 3]]``.

        INPUT:

        - ``rows`` -- boolean (default: ``True``); set to ``True`` if the
          resulting array is by row, otherwise it is by column

        EXAMPLES::

            sage: KRT = crystals.KirillovReshetikhin(['A', 4, 1], 2, 2, model='KR')
            sage: elt = KRT(2, 1, 4, 3)
            sage: elt.to_array()
            [[1, 3], [2, 4]]
            sage: elt.to_array(False)
            [[2, 1], [4, 3]]
        """
        ret_list = []
        h = self.parent()._tableau_height
        s = self.parent()._s
        if rows:
            for i in reversed(range(h)):
                row = []
                for j in range(s):
                    row.append(self[j * h + i])
                ret_list.append(row)
        else:
            for j in range(s):
                col = []
                for i in range(h):
                    col.append(self[j * h + i])
                ret_list.append(col)

        return ret_list

    @cached_method
    def to_tableau(self):
        """
        Return a :class:`Tableau` object of ``self``.

        EXAMPLES::

            sage: KRT = crystals.KirillovReshetikhin(['A', 4, 1], 2, 2, model='KR')
            sage: elt = KRT(2, 1, 4, 3); elt
            [[1, 3], [2, 4]]
            sage: t = elt.to_tableau(); t
            [[1, 3], [2, 4]]
            sage: type(t)
            <class 'sage.combinat.tableau.Tableaux_all_with_category.element_class'>
        """
        return Tableau(self.to_array())

    def pp(self):
        """
        Pretty print ``self``.

        EXAMPLES::

            sage: KRT = crystals.KirillovReshetikhin(['A', 4, 1], 2, 2, model='KR')
            sage: elt = KRT(2, 1, 4, 3); elt
            [[1, 3], [2, 4]]
            sage: elt.pp()
            1  3
            2  4
        """
        self.to_tableau().pp()

    def to_classical_highest_weight(self, index_set=None):
        r"""
        Return the classical highest weight element corresponding to ``self``.

        INPUT:

        - ``index_set`` -- (default: ``None``) return the highest weight
          with respect to the index set; if ``None`` is passed in, then this
          uses the classical index set

        OUTPUT:

        A pair ``[H, f_str]`` where ``H`` is the highest weight element and
        ``f_str`` is a list of `a_i` of `f_{a_i}` needed to reach ``H``.

        EXAMPLES::

            sage: KRTab = crystals.KirillovReshetikhin(['D',4,1], 2, 2, model='KR')
            sage: elt = KRTab(3,2,-1,1); elt
            [[2, 1], [3, -1]]
            sage: elt.to_classical_highest_weight()
            [[[1, 1], [2, -1]], [1, 2]]
        """
        if index_set is None:
            index_set = self.parent()._cartan_type.classical().index_set()
        for i in index_set:
            next = self.e(i)
            if next is not None:
                hw = next.to_classical_highest_weight(index_set=index_set)
                return [hw[0], [i] + hw[1]]
        return [self, []]

    def weight(self):
        """
        Return the weight of ``self``.

        EXAMPLES::

            sage: KR = crystals.KirillovReshetikhin(['D',4,1], 2, 2, model='KR')
            sage: KR.module_generators[1].weight()
            -2*Lambda[0] + Lambda[2]
        """
        return self.Phi() - self.Epsilon()

    @cached_method
    def classical_weight(self):
        r"""
        Return the classical weight of ``self``.

        EXAMPLES::

            sage: KRT = crystals.KirillovReshetikhin(['D',4,1], 2, 2, model='KR')
            sage: elt = KRT(3,2,-1,1); elt
            [[2, 1], [3, -1]]
            sage: elt.classical_weight()
            (0, 1, 1, 0)
        """
        F = self.cartan_type().classical().root_system()
        if F.ambient_space() is None:
            WLR = F.weight_lattice()
        else:
            WLR = F.ambient_space()
        return sum((self[j].weight() for j in range(len(self))), WLR.zero())

    def e(self, i):
        """
        Perform the action of `e_i` on ``self``.

        .. TODO::

            Implement a direct action of `e_0` without moving to KR crystals.

        EXAMPLES::

            sage: KRT = crystals.KirillovReshetikhin(['D',4,1], 2, 2, model='KR')
            sage: KRT.module_generators[0].e(0)
            [[-2, 1], [-1, -1]]
        """
        if i == self.parent()._cartan_type.special_node():
            ret = self.to_kirillov_reshetikhin_crystal().e0()
            if ret is None:
                return None
            return ret.to_kirillov_reshetikhin_tableau()
        return TensorProductOfRegularCrystalsElement.e(self, i)

    def f(self, i):
        """
        Perform the action of `f_i` on ``self``.

        .. TODO::

            Implement a direct action of `f_0` without moving to KR crystals.

        EXAMPLES::

            sage: KRT = crystals.KirillovReshetikhin(['D',4,1], 2, 2, model='KR')
            sage: KRT.module_generators[0].f(0)
            [[1, 1], [2, -1]]
        """
        if i == self.parent()._cartan_type.special_node():
            ret = self.to_kirillov_reshetikhin_crystal().f0()
            if ret is None:
                return None
            return ret.to_kirillov_reshetikhin_tableau()
        return TensorProductOfRegularCrystalsElement.f(self, i)

    def epsilon(self, i):
        r"""
        Compute `\varepsilon_i` of ``self``.

        .. TODO::

            Implement a direct action of `\varepsilon_0` without moving to
            KR crystals.

        EXAMPLES::

            sage: KRT = crystals.KirillovReshetikhin(['D',4,1], 2, 2, model='KR')
            sage: KRT.module_generators[0].epsilon(0)
            2
        """
        if i == self.parent()._cartan_type.special_node():
            return self.to_kirillov_reshetikhin_crystal().epsilon0()
        return TensorProductOfRegularCrystalsElement.epsilon(self, i)

    def phi(self, i):
        r"""
        Compute `\varphi_i` of ``self``.

        .. TODO::

            Compute `\varphi_0` without moving to KR crystals.

        EXAMPLES::

            sage: KRT = crystals.KirillovReshetikhin(['D',4,1], 2, 2, model='KR')
            sage: KRT.module_generators[0].phi(0)
            2
        """
        if i == self.parent()._cartan_type.special_node():
            return self.to_kirillov_reshetikhin_crystal().phi0()
        return TensorProductOfRegularCrystalsElement.phi(self, i)

    def left_split(self):
        r"""
        Return the image of ``self`` under the left column splitting map.

        EXAMPLES::

            sage: KRT = crystals.KirillovReshetikhin(['D',4,1], 2, 3, model='KR')
            sage: mg = KRT.module_generators[1]; mg.pp()
              1 -2  1
              2 -1  2
            sage: ls = mg.left_split(); ls.pp()
              1 (X)  -2  1
              2      -1  2
            sage: ls.parent()
            Tensor product of Kirillov-Reshetikhin tableaux of type ['D', 4, 1] and factor(s) ((2, 1), (2, 2))
        """
        P = self.parent()
        if P._s == 1:
            raise ValueError("cannot split a single column")
        from sage.combinat.rigged_configurations.tensor_product_kr_tableaux import \
            TensorProductOfKirillovReshetikhinTableaux
        r = P._r
        TP = TensorProductOfKirillovReshetikhinTableaux(P._cartan_type, [[r, 1], [r, P._s - 1]])
        lf = TP.crystals[0](*(self[:r]))
        rf = TP.crystals[1](*(self[r:]))
        return TP(lf, rf)

    def right_split(self):
        r"""
        Return the image of ``self`` under the right column splitting map.

        Let `\ast` denote the :meth:`Lusztig involution<lusztig_involution>`,
        and `\mathrm{ls}` as the :meth:`left splitting map<left_split>`.
        The right splitting map is defined as
        `\mathrm{rs} := \ast \circ \mathrm{ls} \circ \ast`.

        EXAMPLES::

            sage: KRT = crystals.KirillovReshetikhin(['D',4,1], 2, 3, model='KR')
            sage: mg = KRT.module_generators[1]; mg.pp()
              1 -2  1
              2 -1  2
            sage: ls = mg.right_split(); ls.pp()
             -2  1 (X)   1
             -1  2       2
            sage: ls.parent()
            Tensor product of Kirillov-Reshetikhin tableaux of type ['D', 4, 1] and factor(s) ((2, 2), (2, 1))
        """
        return self.lusztig_involution().left_split().lusztig_involution()


KirillovReshetikhinTableaux.Element = KirillovReshetikhinTableauxElement


class KRTableauxSpinElement(KirillovReshetikhinTableauxElement):
    r"""
    Kirillov-Reshetikhin tableau for spinors.

    Here we are in the embedding `B(\Lambda_n) \hookrightarrow
    B(2 \Lambda_n)`, so `e_i` and `f_i` act by `e_i^2` and `f_i^2`
    respectively for all `i \neq 0`. We do this so our columns are full
    width (as opposed to half width and/or uses a `\pm` representation).
    """

    def e(self, i):
        r"""
        Calculate the action of `e_i` on ``self``.

        EXAMPLES::

            sage: KRT = crystals.KirillovReshetikhin(['D',4,1], 4, 1, model='KR')
            sage: KRT(-1, -4, 3, 2).e(1)
            [[1], [3], [-4], [-2]]
            sage: KRT(-1, -4, 3, 2).e(3)
        """
        if i == self.parent()._cartan_type.special_node():
            # Only need to do it once since we pull to the KR crystal
            return KirillovReshetikhinTableauxElement.e(self, i)

        half = KirillovReshetikhinTableauxElement.e(self, i)
        if half is None:
            return None
        return KirillovReshetikhinTableauxElement.e(half, i)

    def f(self, i):
        r"""
        Calculate the action of `f_i` on ``self``.

        EXAMPLES::

            sage: KRT = crystals.KirillovReshetikhin(['D',4,1], 4, 1, model='KR')
            sage: KRT(-1, -4, 3, 2).f(1)
            sage: KRT(-1, -4, 3, 2).f(3)
            [[2], [4], [-3], [-1]]
        """
        if i == self.parent()._cartan_type.special_node():
            # Only need to do it once since we pull to the KR crystal
            return KirillovReshetikhinTableauxElement.f(self, i)

        half = KirillovReshetikhinTableauxElement.f(self, i)
        if half is None:
            return None

        return KirillovReshetikhinTableauxElement.f(half, i)

    def epsilon(self, i):
        r"""
        Compute `\varepsilon_i` of ``self``.

        EXAMPLES::

            sage: KRT = crystals.KirillovReshetikhin(['D',4,1], 4, 1, model='KR')
            sage: KRT(-1, -4, 3, 2).epsilon(1)
            1
            sage: KRT(-1, -4, 3, 2).epsilon(3)
            0
        """
        if i == self.parent()._cartan_type.special_node():
            # Don't need to half it since we pull to the KR crystal
            return KirillovReshetikhinTableauxElement.epsilon(self, i)
        return KirillovReshetikhinTableauxElement.epsilon(self, i) // 2

    def phi(self, i):
        r"""
        Compute `\varphi_i` of ``self``.

        EXAMPLES::

            sage: KRT = crystals.KirillovReshetikhin(['D',4,1], 4, 1, model='KR')
            sage: KRT(-1, -4, 3, 2).phi(1)
            0
            sage: KRT(-1, -4, 3, 2).phi(3)
            1
        """
        if i == self.parent()._cartan_type.special_node():
            # Don't need to half it since we pull to the KR crystal
            return KirillovReshetikhinTableauxElement.phi(self, i)
        return KirillovReshetikhinTableauxElement.phi(self, i) // 2

    @cached_method
    def to_array(self, rows=True):
        r"""
        Return a 2-dimensional array representation of this
        Kirillov-Reshetikhin element.

        If the output is in rows, then it outputs the top row first (in the
        English convention) from left to right.

        For example: if the reading word is `[2, 1, 4, 3]`, so as a
        `2 \times 2` tableau::

            1 3
            2 4

        we output ``[[1, 3], [2, 4]]``.

        If the output is in columns, then it outputs the leftmost column first
        with the bottom element first. In other words this parses the reading
        word into its columns.

        Continuing with the previous example, the output would be
        ``[[2, 1], [4, 3]]``.

        INPUT:

        - ``rows`` -- boolean (default: ``True``); set to ``True`` if the
          resulting array is by row, otherwise it is by column

        EXAMPLES::

            sage: KRT = crystals.KirillovReshetikhin(['D', 4, 1], 4, 3, model='KR')
            sage: elt = KRT(-3,-4,2,1,-3,-4,2,1,-2,-4,3,1)
            sage: elt.to_array()
            [[1, 1, 1], [2, 2, 3], [-4, -4, -4], [-3, -3, -2]]
            sage: elt.to_array(False)
            [[-3, -4, 2, 1], [-3, -4, 2, 1], [-2, -4, 3, 1]]
        """
        ret_list = []
        h = self.parent()._cartan_type.classical().rank()
        s = self.parent()._s
        if rows:
            for i in reversed(range(h)):
                row = []
                for j in range(s):
                    row.append(self[j * h + i])
                ret_list.append(row)
        else:
            for j in range(s):
                col = []
                for i in range(h):
                    col.append(self[j * h + i])
                ret_list.append(col)

        return ret_list

    def left_split(self):
        """
        Return the image of ``self`` under the left column splitting map.

        EXAMPLES::

            sage: KRT = crystals.KirillovReshetikhin(['D', 4, 1], 4, 3, model='KR')
            sage: elt = KRT(-3,-4,2,1,-3,-4,2,1,-2,-4,3,1); elt.pp()
              1  1  1
              2  2  3
             -4 -4 -4
             -3 -3 -2
            sage: elt.left_split().pp()
              1 (X)   1  1
              2       2  3
             -4      -4 -4
             -3      -3 -2
        """
        P = self.parent()
        if P._s == 1:
            raise ValueError("cannot split a single column")
        from sage.combinat.rigged_configurations.tensor_product_kr_tableaux import \
            TensorProductOfKirillovReshetikhinTableaux
        h = P._cartan_type.classical().rank()
        TP = TensorProductOfKirillovReshetikhinTableaux(P._cartan_type, [[P._r, 1], [P._r, P._s - 1]])
        lf = TP.crystals[0](*(self[:h]))
        rf = TP.crystals[1](*(self[h:]))
        return TP(lf, rf)

    # FIXME: This is a copy of the above classical weight, and cached_method
    #   overwrites this method if it is called via super.
    @cached_method
    def classical_weight(self):
        r"""
        Return the classical weight of ``self``.

        EXAMPLES::

            sage: KRT = crystals.KirillovReshetikhin(['D', 4, 1], 4, 1, model='KR')
            sage: KRT.module_generators[0].classical_weight()
            (1/2, 1/2, 1/2, 1/2)
        """
        F = self.cartan_type().classical().root_system()
        if F.ambient_space() is None:
            WLR = F.weight_lattice()
        else:
            WLR = F.ambient_space()
        return sum((self[j].weight() for j in range(len(self))), WLR.zero()) / 2


KRTableauxBn.Element = KRTableauxSpinElement
KRTableauxSpin.Element = KRTableauxSpinElement


class KRTableauxDTwistedSpin(KRTableauxRectangle):
    r"""
    Kirillov-Reshetikhin tableaux `B^{r,s}` of type `D_n^{(2)}` with `r = n`.

    EXAMPLES::

        sage: KRT = crystals.KirillovReshetikhin(['D', 4, 2], 1, 1, model='KR')
        sage: KRT.cardinality()
        8
        sage: KRC = crystals.KirillovReshetikhin(['D', 4, 2], 1, 1, model='KN')
        sage: KRT.cardinality() == KRC.cardinality()
        True
    """
    Element = KRTableauxSpinElement


class KRTableauxTypeFromRCElement(KirillovReshetikhinTableauxElement):
    r"""
    A Kirillov-Reshetikhin tableau constructed from rigged configurations
    under the bijection `\Phi`.
    """

    def e(self, i):
        """
        Perform the action of `e_i` on ``self``.

        .. TODO::

            Implement a direct action of `e_0` without moving to
            rigged configurations.

        EXAMPLES::

            sage: KRT = crystals.KirillovReshetikhin(['D',4,3], 2, 1, model='KR')
            sage: KRT.module_generators[0].e(0)
            [[2], [E]]
        """
        if i == self.parent().cartan_type().special_node():
            P = self.parent()
            from sage.combinat.rigged_configurations.tensor_product_kr_tableaux import TensorProductOfKirillovReshetikhinTableaux
            K = TensorProductOfKirillovReshetikhinTableaux(P.cartan_type(), [[2, P.s()]])
            ret = K(self).to_rigged_configuration()
            RC = ret.parent()
            ret = ret.to_virtual_configuration().e(0)
            if ret is None:
                return None
            ret = RC.from_virtual(ret)
            return ret.to_tensor_product_of_kirillov_reshetikhin_tableaux()[0]
        return TensorProductOfRegularCrystalsElement.e(self, i)

    def f(self, i):
        """
        Perform the action of `f_i` on ``self``.

        .. TODO::

            Implement a direct action of `f_0` without moving to
            rigged configurations.

        EXAMPLES::

            sage: KRT = crystals.KirillovReshetikhin(['D',4,3], 2, 1, model='KR')
            sage: KRT.module_generators[0].f(0)
            sage: KRT.module_generators[3].f(0)
            [[1], [0]]
        """
        if i == self.parent().cartan_type().special_node():
            P = self.parent()
            from sage.combinat.rigged_configurations.tensor_product_kr_tableaux import TensorProductOfKirillovReshetikhinTableaux
            K = TensorProductOfKirillovReshetikhinTableaux(P.cartan_type(), [[2, P.s()]])
            ret = K(self).to_rigged_configuration()
            RC = ret.parent()
            ret = ret.to_virtual_configuration().f(0)
            if ret is None:
                return None
            ret = RC.from_virtual(ret)
            return ret.to_tensor_product_of_kirillov_reshetikhin_tableaux()[0]
        return TensorProductOfRegularCrystalsElement.f(self, i)

    def epsilon(self, i):
        r"""
        Compute `\varepsilon_i` of ``self``.

        .. TODO::

            Implement a direct action of `\epsilon_0` without moving to
            KR crystals.

        EXAMPLES::

            sage: KRT = crystals.KirillovReshetikhin(['D',4,3], 2, 2, model='KR')
            sage: KRT.module_generators[0].epsilon(0)
            6
        """
        if i == self.parent().cartan_type().special_node():
            P = self.parent()
            from sage.combinat.rigged_configurations.tensor_product_kr_tableaux import TensorProductOfKirillovReshetikhinTableaux
            K = TensorProductOfKirillovReshetikhinTableaux(P.cartan_type(), [[2, P.s()]])
            rc = K(self).to_rigged_configuration().to_virtual_configuration()
            return rc.epsilon(0)
        return TensorProductOfRegularCrystalsElement.epsilon(self, i)

    def phi(self, i):
        r"""
        Compute `\varphi_i` of ``self``.

        .. TODO::

            Compute `\phi_0` without moving to KR crystals.

        EXAMPLES::

            sage: KRT = crystals.KirillovReshetikhin(['D',4,3], 2, 2, model='KR')
            sage: KRT.module_generators[0].phi(0)
            0
        """
        if i == self.parent().cartan_type().special_node():
            P = self.parent()
            from sage.combinat.rigged_configurations.tensor_product_kr_tableaux import TensorProductOfKirillovReshetikhinTableaux
            K = TensorProductOfKirillovReshetikhinTableaux(P.cartan_type(), [[2, P.s()]])
            rc = K(self).to_rigged_configuration().to_virtual_configuration()
            return rc.phi(0)
        return TensorProductOfRegularCrystalsElement.phi(self, i)


class KRTableauxTypeFromRC(KirillovReshetikhinTableaux):
    r"""
    Kirillov-Reshetikhin tableaux `B^{r,s}` constructed from rigged
    configurations under the bijection `\Phi`.

    .. WARNING::

        The Kashiwara-Nakashima version is not implemented due to the
        non-trivial multiplicities of classical components, so
        :meth:`classical_decomposition` does not work.
    """

    def __init__(self, cartan_type, r, s):
        r"""
        Initialize ``self``.

        EXAMPLES::

            sage: KRT = crystals.KirillovReshetikhin(['D', 4, 3], 2, 1, model='KR')
            sage: TestSuite(KRT).run() # long time
        """
        # We must modify the constructor of KirillovReshetikhinTableaux
        self._r = r
        self._s = s
        self._cartan_type = cartan_type
        Parent.__init__(self, category=KirillovReshetikhinCrystals())
        self.letters = CrystalOfLetters(cartan_type.classical())

    @lazy_attribute
    def module_generators(self):
        """
        The module generators of ``self``.

        EXAMPLES::

            sage: KRT = crystals.KirillovReshetikhin(['D',4,3], 2, 1, model='KR')
            sage: KRT.module_generators
            ([[1], [2]], [[1], [0]], [[1], [E]], [[E], [E]])
        """
        return self._build_module_generators()

    def _build_module_generators(self):
        r"""
        Return the module generators of ``self``.

        EXAMPLES::

            sage: KRT = crystals.KirillovReshetikhin(['D',4,3], 2, 1, model='KR')
            sage: KRT._build_module_generators()
            ([[1], [2]], [[1], [0]], [[1], [E]], [[E], [E]])
        """
        from sage.combinat.rigged_configurations.rigged_configurations import RiggedConfigurations
        RC = RiggedConfigurations(self._cartan_type, [[self._r, self._s]])
        return tuple(mg.to_tensor_product_of_kirillov_reshetikhin_tableaux()[0]
                     for mg in RC.module_generators)

    @lazy_attribute
    def _tableau_height(self):
        """
        The height of the tableaux in ``self``.

        EXAMPLES::

            sage: ct = CartanType(['E',6,1])
            sage: [crystals.KirillovReshetikhin(ct, r, 1, model='KR')._tableau_height
            ....:  for r in ct.classical().index_set()]
            [1, 3, 2, 3, 4, 2]
        """
        if self._cartan_type.type() == 'E':
            if self._cartan_type.classical().rank() == 6:
                #       6   2 - 5
                #      /   /
                # 0 - 1 - 3 - 4
                if self._r == 1:
                    return 1
                if self._r == [3, 6]:
                    return 2
                if self._r in [2, 4]:
                    return 3
                if self._r == 5:
                    return 4
            if self._cartan_type.classical().rank() == 7:
                #     1-2-3
                #    /
                # 0-7-6-5-4
                if self._r <= 3:
                    return self._r + 1
                return 8 - self._r
            if self._cartan_type.classical().rank() == 8:
                #     1-2-3
                #    /
                # 0-8-7-6-5-4
                if self._r <= 3:
                    return self._r + 1
                return 9 - self._r
        if not self._cartan_type.is_untwisted_affine():
            if self._cartan_type.dual().type() == 'G':
                return self._r
        return len(self.module_generators[0]) // self._s

    Element = KRTableauxTypeFromRCElement
